/*

 Author:	Matt Fausey & Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				JM_ASSIGN.C												*/
/*																		*/
/*																		*/
/************************************************************************/


/*****************/
/* Include Files */
/*****************/

#include "all.h"

/********************/
/* Global Variables */
/********************/

int logical_cpus[ MAX_SDF_CPUS ];	/* pointers to real cpus in sdf_cpu array */
unsigned long *labeled_cpus;		/* set of real cpus that have pointers
										in logical_cpus array */

static int total_processes = 0;
static int total_cpus = 0;

int cpu_equivalence();

/******************************************************************/
/*
	FUNCTION NAME:

		assign_cpus

	DESCRIPTION:

		Main routine for assigning processes to cpus

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
assign_cpus()
{
	initialize_process_info();
	check_for_max_processes();
	build_initial_class_sets();
	build_final_class_sets();
	do_build_maps();
}

/******************************************************************/
/*
	FUNCTION NAME:

		initialize_logical_cpus

	DESCRIPTION:

		Initialize logical cpus to -1

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
initialize_logical_cpus()
{
	int cpu;

	for( cpu = 0; cpu < MAX_SDF_CPUS; cpu++ )
		logical_cpus[cpu] = -1;

	labeled_cpus = create_empty_set( MAX_SDF_CPUS );
}

/******************************************************************/
/*
	FUNCTION NAME:

		create_logical_cpu

	DESCRIPTION:

		Create a logical cpu from SDF file

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
create_logical_cpu(cpu)
int cpu;
{
	char errmsg[80];

	if( cpu < 0 )
	{
		sprintf(errmsg,"Warning: Logical cpu %d not in range 1-256",cpu+1);
		printerr(FACILITY,WARNING_MSG,errmsg,NULL,67);
		add_syntax_error();
	}
	else if( logical_cpus[ cpu ] != -1 )
	{
		sprintf(errmsg,"Warning: Multiple declaration of logical cpu %d",cpu+1);
		printerr(FACILITY,WARNING_MSG,errmsg,NULL,64);
		add_syntax_error();
	}
	else
	{
		logical_cpus[ cpu ] = total_cpus++;
		add_to_set( total_cpus, labeled_cpus );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		initialize_process_info

	DESCRIPTION:

		Loop through classes and do the following:

		1.	If a class has logical cpus (e.g. a line: "x processes on cpu y"
			in the jdf file) then call "assign_processes_to_logical_cpus"
			to assign its processes.

		2.	If a class does not have logical cpus then calculate number of
			cpus, processes per cpu and number of processes.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
initialize_process_info()
{
	int i, j;
	char errmsg[80],action[80];

	for( i = 1; i <= MAX_CLASSES; i++ )
	{
		if( class[i].valid )
		{
			if( class[i].logical_assignment )
			{
				assign_processes_to_logical_cpus( i );
			}
			else if( class[i].nproc != AS_MANY_AS_POSSIBLE )
			{
				calculate_process_info( i );

/* If we requested specific CPUs, check to see if we requested enough */
/*
				printf("Class %d: nspec = %d   nproc = %d  ncpu = %d\n",
					i,class[i].nspec,class[i].nproc,class[i].ncpu);
*/
				if( class[i].nspec && (class[i].ncpu > class[i].nspec) )
				{
					sprintf(errmsg,"Too few CPUs requested for class %d",i);
					sprintf(action,
		   				"Need to specify %d more specific CPUs",
						class[i].ncpu - class[i].nspec);
					printerr(FACILITY,FATAL_MSG,errmsg,action,1);
					exit(0);
				}

/* If we requested too many specific CPUs, do not use them all */

				if( class[i].nspec > class[i].ncpu )
				{
					sprintf(errmsg,
						"Warning: More specific CPUs specified for class %d than needed",i);
					printerr(FACILITY,WARNING_MSG,errmsg,NULL,2);
				}
				for( j = class[i].ncpu; j < class[i].nspec; j++ )
					remove_from_set(i,sdf_cpu[class[i].spec_cpu[j]].class);
				if( class[i].nspec > class[i].ncpu )
					class[i].nspec = class[i].ncpu;
			}
/* Handle AMAP */
			else
			{
				if( class[i].per_cpu == ALL_ON_ONE_CPU )
					class[i].per_cpu = MAX_PROCESSES;
				else if( !class[i].per_cpu )
					class[i].per_cpu = 1;

				if( class[i].nspec )
				{
					sprintf(errmsg,
			  			"Cannot assign specific CPUs while \
						requesting as many as possible processes (class = %d)",i);
					printerr(FACILITY,FATAL_MSG,errmsg,NULL,3);
					exit(0);
				}
			}
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		calculate_process_info

	DESCRIPTION:

		Calculate number of processes, processes per cpu, and number of cpus.
		This function is called for all classes except those with logical
		cpus and those with AMAP number of processes.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
calculate_process_info( i )
int i;
{
	if( class[i].per_cpu == ALL_ON_ONE_CPU )
	{
		if( !class[i].nproc )
			class[i].nproc = 1;
   		class[i].per_cpu = class[i].nproc;
		class[i].ncpu = 1;
	}

	if( !class[i].ncpu )
	{
		if( class[i].nproc && class[i].per_cpu )
		{
			class[i].ncpu =
				class[i].nproc / class[i].per_cpu;
			class[i].ncpu +=
				(class[i].nproc % class[i].per_cpu) ? 1 : 0;
		}
		else
			class[i].ncpu = 1;
	}

	if( !class[i].nproc )
	{
		if( class[i].ncpu && class[i].per_cpu )
			class[i].nproc = class[i].ncpu * class[i].per_cpu;
		else
			class[i].nproc = 1;
	}

	if( !class[i].per_cpu )
	{
		if( class[i].nproc && class[i].ncpu )
		{
			class[i].per_cpu =
				class[i].nproc / class[i].ncpu;
				/*
			class[i].per_cpu +=
				(class[i].nproc % class[i].ncpu) ? 1 : 0;
				*/
		}
		else
			class[i].per_cpu = 1;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		assign_processes_to_logical_cpus

	DESCRIPTION:

		Assign processes to cpus in classes that use logical cpus.
		Search the class's logical_processes list to see how many
		processes to put on each cpu. Make entries in the logical_cpus
		list. Called only by classes with logical cpus.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
assign_processes_to_logical_cpus( i )
int i;
{
	int j, k;
	char errmsg[80];
	int size1, size2;
	unsigned long *s;

	class[i].nproc = 0;
	class[i].ncpu = 0;
	for( j = 0; j < MAX_SDF_CPUS; j++ )
	{
		if( class[i].logical_processes[j] )
		{
			k = 0;
			if( logical_cpus[j] == -1 )
			{
				sprintf( errmsg,
					"Cannot access logical cpu %d in class %d", j + 1, i );
				printerr( FACILITY,FATAL_MSG, errmsg, NULL, 65 );
				exit(0);
			}

			size1 = set_size( sdf_cpu[ logical_cpus[j] ].class, MAX_CLASSES );
			s = set_intersect( sdf_cpu[ logical_cpus[j] ].class,
				class[i].share, MAX_CLASSES );
			size2 = set_size( s, MAX_CLASSES );
			/*
			if( ( size1 != size2 ) ||
				!cpu_equivalence( logical_cpus[j], i ) )
			*/
			if( size1 != size2 )
			{
				sprintf( errmsg,
					"Logical cpu %d cannot be used in class %d", j + 1, i );
				printerr( FACILITY,FATAL_MSG, errmsg, NULL, 66 );
				exit(0);
			}

			class[i].ncpu++;
			class[i].nproc += class[i].logical_processes[j];
			add_to_set( k + 1, class[i].cpu);
			sdf_cpu[ logical_cpus[j] ].nprocs += class[i].logical_processes[j];
			total_processes += class[i].logical_processes[j];
			add_to_set( i, sdf_cpu[ logical_cpus[j] ].class );
		}		
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		cpu_equivalence

	DESCRIPTION:

		Check to see if a cpu matches those requested by a class.
		Compare cpu_type and task.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
int
cpu_equivalence( cpu, cclass )
int cpu, cclass;
{
	int value;

	if
	(
		( sdf_cpu[ cpu ].cpu_type == class[ cclass ].cpu_type ) &&
		(
			( sdf_cpu[ cpu ].task ==  class[ cclass ].task ) ||
			( sdf_cpu[ cpu ].task == TASK_ANY ) ||
			( class[ cclass ].task == TASK_ANY )
		)
	)
		value = 1;
	else
		value = 0;

	return( value );
}

/******************************************************************/
/*
	FUNCTION NAME:

		check_for_max_processes

	DESCRIPTION:

		Check to see if we've gone over the limit on the total
		number of processes. This is called before AMAP classes
		are handled so it is possible that there still will be too
		many processes.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
check_for_max_processes()
{
	int i;
	char errmsg[80];
	int total;

	total = 0;
	for( i = 1; i <= MAX_CLASSES; i++ )
		if(class[i].valid && ( class[i].nproc != AS_MANY_AS_POSSIBLE ) )
		{
			total += class[i].nproc;
		}
	if(total > MAX_PROCESSES)
	{
		sprintf(errmsg,
			"Too many processes have been requested, maximum is %d",
			MAX_PROCESSES);
		printerr(FACILITY,FATAL_MSG,errmsg,NULL,5);
		exit(0);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		build_initial_class_sets

	DESCRIPTION:

		For each class, build a set of possible CPUs.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
build_initial_class_sets()
{
	int i;
	int size1;
	char errmsg[80], action[80];

	for( i = 1; i <= MAX_CLASSES; i++ )
	{
		if( class[i].valid && !class[i].nspec && !class[i].logical_assignment ) 
		{
			build_set( i );
			if( class[i].nproc != AS_MANY_AS_POSSIBLE )
			{
				size1 = set_size( class[i].cpu,MAX_SDF_CPUS );
				if( class[i].ncpu > size1 )
				{
					sprintf(errmsg,
			   			"Not enough CPUs in cps_sdf file for class %d",i);
					sprintf(action,
						"cps_sdf file needs %d more CPUs of the correct type",
			   			class[i].ncpu - size1);
					printerr(FACILITY,FATAL_MSG,errmsg,action,4);
					exit(0);
				}
			}
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		build_set

	DESCRIPTION:

		Do the actual set building for the subroutine above.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
build_set( i )
int i;
{
	int j;
	int size1,size2;
	unsigned long *s;

/* add the set of available cpus for each class */
/* checking for specific cpus, cpu type, and cpu task */
	for( j = 0; j < num_sdf_cpus; j++)
	{
		if( cpu_equivalence( j, i ) )
		{
/* Check for cpus that have been chosen specifically by a class */
/* size1 = 0 means that the cpu is not a specific cpu in any class */
			size1 = set_size( sdf_cpu[j].class, MAX_CLASSES );
			if( size1 == 0 )
			{
				add_to_set( j + 1, class[i].cpu );
			}
/* Intersect specific cpus with share classes to see if the cpu can be shared */
/* and thereby added to the class cpu list */
			else
			{
				s = set_intersect( sdf_cpu[j].class,
					class[i].share,MAX_CLASSES );
				size2 = set_size( s, MAX_CLASSES );
				if( size1 == size2 )
					add_to_set( j + 1, class[i].cpu );
			}
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		build_final_class_sets

	DESCRIPTION:

		Loop through the classes and assign cpus to classes. Classes
		with logical cpus are ignored since they have already been
		assigned. Assign to classes in order of number of share classes.
		(e.g. A class that does not share will have cpus assigned first)
		After all these are done, do all AMAP classes.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
build_final_class_sets()
{
	int i, j;
	char errmsg[80];
	int amap[ MAX_CLASSES ];

	for( i = 0; i < MAX_CLASSES; i++ )
		amap[ i ] = 0;

/* loop on number of classes shared */
	for( i = 1; i <= MAX_CLASSES; i++ )
	{
/* loop on the classes themselves */
		for( j = 1; j <= MAX_CLASSES; j++ )
		{
			if( class[ j ].valid && !class[ j ].logical_assignment &&
				( class[ j ].nproc != AS_MANY_AS_POSSIBLE ) &&
				( set_size( class[ j ].share, MAX_CLASSES ) == i ) )
			{
				assign_cpus_to_class( j );
			}
			else if( class[ j ].valid &&
				( class[ j ].nproc == AS_MANY_AS_POSSIBLE ) )
			{
				amap[ j - 1 ] = 1;
			}
		}
	}
/* Handle AMAP last */
	for( i = 1; i <= MAX_CLASSES; i++ )
	{
		if( amap[ i - 1 ] )
		{
			class[ i ].ncpu = set_size( class[ i ].cpu, MAX_SDF_CPUS );
			class[ i ].nproc = class[ i ].ncpu * class[ i ].per_cpu;
			assign_cpus_to_class( i );
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		assign_cpus_to_class

	DESCRIPTION:

		Do the actual assignment of cpus to classes. Classes with logical
		cpus do not call this routine since they have cpus assigned to them
		already. AMAP classes do go through here and there is a process
		limit check in this routine. Logical cpus are created for all real
		cpus in the job so that the info is stored consistently for later
		retrieval. For each cpu in the set of possible CPUs, the CPU record
		is updated. For each class that cannot share with this class, remove
		the CPU from its set.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
assign_cpus_to_class( c )
int c;
{
	int *x1, x[ MAX_SDF_CPUS ];
	int i, j;
	int processes_left;
	int cpus_left;
	int cpus_available;
	int processes_assigned;
	char errmsg[80];

/* Get a list of CPUs */

	x1 = set_query( class[c].cpu, MAX_SDF_CPUS );
	i = 0;
	while( x1[i] != 0 )
	{
		x[i] = x1[i];
		i++;
	}
	x[i] = 0;

/* sort list in order of least loaded first */
	sort_cpu_list( x );

	cpus_left = class[c].ncpu;
	processes_left = class[c].nproc;

	cpus_available = set_size( class[c].cpu, MAX_SDF_CPUS );
	if( cpus_left > cpus_available || cpus_available == 0 )
	{
		sprintf(errmsg,
  			"Impossible to assign enough CPUs for class %d", c);
		printerr(FACILITY,FATAL_MSG,errmsg,NULL,6);
		exit(0);
	}
	if( total_processes >= MAX_PROCESSES )
	{
		sprintf(errmsg,
			"Too many processes have been requested, maximum is %d",
			MAX_PROCESSES);
		printerr(FACILITY,FATAL_MSG,errmsg,NULL,5);
		exit(0);
	}

/* Do each CPU */

	i = 0;
	while( x[i] && cpus_left )
	{

/* Calculate how many processes will run on this cpu */
		if( processes_left < cpus_left )
			processes_assigned = class[c].per_cpu;
		else
			processes_assigned = processes_left / cpus_left;

		cpus_left = cpus_left--;
		processes_left -= processes_assigned;

		if( ( processes_assigned + total_processes ) > MAX_PROCESSES )
		{
			processes_assigned = MAX_PROCESSES - total_processes;
			processes_left = 0;
		}

		total_processes += processes_assigned;

		sdf_cpu[ x[i] - 1 ].nprocs += processes_assigned;

/* create a logical cpu for the real cpu so we can utilize this info later */
/* every real cpu will therefore have a logical one assocated with it */
		if( !set_member( x[i], labeled_cpus ) )
			generate_logical_cpu( c, processes_assigned, x[i] );
		else
			assign_logical_cpu( c, processes_assigned, x[i] );

/* Add the class to the CPU's set of classes */

		add_to_set( c, sdf_cpu[ x[i] - 1 ].class);

/* For each class that cannot share with this class, remove this cpu from
   the set of possible CPUs */

		for( j = 1; j <= MAX_CLASSES; j++ )
			if( class[j].valid && !class[j].nspec &&
				!set_member( c, class[j].share ) )
			{
				remove_from_set( x[i], class[j].cpu );
			}

		i++;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		sort_cpu_list

	DESCRIPTION:
  
		Sort cpu list in order of least processes per cpu. This routine
		will help balance the load on the job's cpus.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

/******************************************************************/
sort_cpu_list( x )
int x[ MAX_SDF_CPUS ];
{
	int i;
	int y[ MAX_SDF_CPUS ];
	int procs[ MAX_SDF_CPUS ];
	int lowprocs;
	int listlength;
	int count;

/* determine how many cpus are in list by searching for a "0" */
/* listlength = number of cpus */
	listlength = 0;
	while( x[ listlength ] != 0 )
		listlength++;

/* copy nprocs to a new array so that we don't modify the original */
/* just to make the sort loop simpler */
	for( i = 0; i < MAX_SDF_CPUS; i++ )
		procs[ i ] = sdf_cpu[ i ].nprocs;

	count = 0;
	while( count < listlength )
	{
		lowprocs = MAX_PROCESSES + 1;
		for( i = 0; i < listlength; i++ )
		{
			if( procs[ x[ i ] - 1 ] < lowprocs )
			{
				y[ count ] = x[ i ];
				lowprocs = procs[ x[ i ] - 1 ];
			}
		}
/* set procs for this cpu to be large so that it's not picked more than once */
		procs[ y[ count ] - 1 ] = MAX_PROCESSES + 1;
		count++;
	}

/* move the values to the array that was passed and return */
	for( i = 0; i < listlength; i++ )
		x[ i ] = y[ i ];
}

/******************************************************************/
/*
	FUNCTION NAME:

		generate_logical_cpu

	DESCRIPTION:

		For those cpus that have not been explicitly defined as having
		a logical mate, create one for them. I found that having a logical
		cpu for all cpus made it easier to later load data into the user
		maps and might be useful for future enhancements to the program.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
generate_logical_cpu( cclass, processes, cpu )
int cclass, processes, cpu;
{
	int i;

	for( i = 0; i < MAX_SDF_CPUS; i++ )
	{
		if( logical_cpus[ i ] == -1 )
		{
			logical_cpus[ i ] = cpu - 1;
			assign_processes_to_cpu( cclass, processes, i + 1, i + 1 );
			add_to_set( cpu, labeled_cpus );
			break;
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		assign_logical_cpu

	DESCRIPTION:

		If a cpu already has a logical mate, we do not need to create
		one, merely assign the one that already exists.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
assign_logical_cpu( cclass, processes, cpu )
{
	int i;

	for( i = 0; i < MAX_SDF_CPUS; i++ )
	{
		if( logical_cpus[ i ] == cpu - 1 )
		{
			assign_processes_to_cpu( cclass, processes, i + 1, i + 1 );
			break;
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		do_build_maps

	DESCRIPTION:

		Sets up the looping for storing data in the user maps.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
do_build_maps()
{
	int c, cpu;

	total_processes = 0;
	for( c = 1; c <= MAX_CLASSES; c++ )
		for( cpu = 0; cpu < MAX_SDF_CPUS; cpu++ )
			if( class[ c ].logical_processes[ cpu ] != 0 )
				build_maps( c, cpu );
}

/******************************************************************/
/*
	FUNCTION NAME:

		build_maps

	DESCRIPTION:

		Load data in user maps and terminate this whole section.

	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
/*
	CPUs in class[].cpu are now the ones to be assigned
	Build the user & shm maps
*/
build_maps( cclass, cpu )
int cclass, cpu;
{
	int k;
	int pleft;

	pleft = class[ cclass ].logical_processes[ cpu ];
	cpu = logical_cpus[ cpu ];

	while( pleft > 0 )
	{
		usr_map[total_processes].class = cclass;
		usr_map[total_processes].cpu = cpu;
		usr_map[total_processes].status = 0;
		usr_map[total_processes].sock = 0;

		k = 0;
		while((k < num_shm_procs) && (shm_map[k].cpu != cpu))
			k++;
		if(k == num_shm_procs)
		{
			shm_map[k].proc = create_empty_set(MAX_PROCESSES);
			shm_map[k].cpu = cpu;
			shm_map[k].status = 0;
			shm_map[k].sock = 0;
			num_shm_procs++;
		}

		add_to_set( total_processes + 1, shm_map[k].proc );
		usr_map[ total_processes ].shm = k;

		add_to_set( total_processes + 1, sdf_cpu[ cpu ].procs);
		add_to_set( total_processes + 1, class[ cclass ].sop);
		
		total_processes++;
		pleft--;
	}
	num_usr_procs = total_processes;
}

/******************************************************************/
/*
	FUNCTION NAME:

		assign_processes_to_cpu

	DESCRIPTION:


	FORMAL PARAMETERS:

		[@description or none@]

	FUNCTION VALUE:

		[@description or none@]

								  */
/******************************************************************/
assign_processes_to_cpu( cclass, processes, cpu1, cpu2 )
int cclass, processes, cpu1, cpu2;
{
	char errmsg[80];
	int cpu;

	for( cpu = cpu1; cpu <= cpu2; cpu++ )
	{
		if( ( cpu > MAX_SDF_CPUS ) || ( cpu < 1 ) )
		{
			sprintf(errmsg,"CPU number %d in class %d is invalid,\
			number must be between 1 and %d\n",
				cpu, cclass, MAX_SDF_CPUS );
			printerr(FACILITY,FATAL_MSG,errmsg,NULL,3);
			exit(0);
		}
		class[ cclass ].logical_assignment = 1;
		class[ cclass ].logical_processes[ cpu - 1 ] = processes;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	assign_logical_cpu_to_class

	DESCRIPTION:

	FORMAL PARAMETERS:

	[@description or none@]

	FUNCTION VALUE:

	[@description or none@]

*/
/******************************************************************/
add_logical_cpu_to_class( cclass, cpu )
int cclass, cpu;
{
	char IllegalLogicalCPU[16];
	char errmsg[80];

	if( cpu < 1 )
	{
		sprintf(errmsg,"Logical cpu %d in class %d not in range 1-256",cpu,
			cclass);
		printerr(FACILITY,FATAL_MSG,errmsg,NULL,69);
		exit(0);
	}
	else if( sdf_cpu[ logical_cpus[ cpu - 1 ] ].cpu_name == NULL )
	{
		sprintf( IllegalLogicalCPU, "%s %d", "Logical CPU", cpu );
		add_cpu_to_class( cclass, IllegalLogicalCPU );
	}
	else
	{
		add_cpu_to_class( cclass, sdf_cpu[ logical_cpus[ cpu - 1 ] ].cpu_name );
	}
}
