/* CPS_SYSMON.C V1.2						4 Jan, 1993

 Author:        Bill Meyer
                Fermi National Accelerator Laboratory - Mail Station 369
                P.O Box 500
                Batavia, IL 60510
                Internet: meyer@fnfaus.fnal.gov

 Copyright (c) 1993 Universities Research Association, Inc.
       All Rights Reserved


 This material resulted from work developed under a Government Contract and
 is subject to the following license:  The Government retains a paid-up,
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative
 works, perform publicly and display publicly by or for the Government,
 including the right to distribute to other Government contractors.  Neither
 the United States nor the United States Department of Energy, nor any of
 their employees, makes any warrenty, express or implied, or assumes any
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or
 represents that its use would not infringe privately owned rights.

*/

#include "../inc/types.h"

#ifdef AIX
#include <sys/select.h>
#endif
#include <sys/param.h>
#include <sys/socket.h>
#include <rpc/rpc.h>
#include <rpc/pmap_clnt.h>
#include <rpcsvc/rstat.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <string.h>
#include <arpa/inet.h>
#include <curses.h>

#define DB_GLOBALS 1

#include "../inc/database.h"
#include "../inc/db_functions.h"
#include "../inc/perr.h"

#define FACILITY "SYSMON"
extern struct database db;
extern char *getenv(const char *);

#define P_CPU		0
#define P_PAGE		1
#define P_SWAP		2
#define P_INTR		3
#define P_DISK		4
#define P_CNTXT		5
#define P_LOAD		6
#define P_TPKTS		7
#define P_IPKTS		8
#define P_OPKTS		9
#define P_COLLS		10
#define P_ERRS		11
#define MAXITEMS	12

#ifdef AIX
#define CP_USER 0
#define CP_NICE 1
#define CP_SYS  2
#define CP_IDLE 3
#endif

#define FROM_FILE	1
#define	FROM_PM		2
#define UNDEFINED	3

int input_option = UNDEFINED;

char hostnm[MAXHOSTNAMELEN];
int update = 20, num_display_lines = 0, total_cpus = 0, flag = 0;

struct prod_sys *p;

struct CPULIST {
    char name[81];
    CLIENT *client;
    struct sockaddr_in server_addr;
    struct statstime stats_info, prev_v;
    enum clnt_stat status;
    int sock;
    int data[MAXITEMS];
    struct CPULIST *next;
} *head;

void get_data(void);
void display_data(void);
void get_cpulist_from_file(char *file);
void get_cpulist_from_pm(char *name);
static int compare();
void acpu(char *name);
void rcpu(char *name);
void truncate(char *p);
void destroy_list(void);
void establish_remote_connections(void);
void sigint_handler();


void main(int argc, char **argv)
{
    char buffer[255];
    extern char *optarg;
    extern int optind;
    int i;

    initscr();
    num_display_lines = (LINES-4);
    endwin();

    while((i = getopt(argc, argv, "f:hp:u:")) != -1)
	switch(i) {
            case 'f':
                input_option = FROM_FILE;
                get_cpulist_from_file(optarg);
                break;

	    case 'h':
printf("\nUsage: cps_sysmon -h [-f <filename>][-p <production_system>] [-u <update>]\n\n");
printf("	CPU:	Pecentage of CPU being utilized\n");
printf("	PAGE:	Paging activity in pages per second\n");
printf("	SWAP:	Jobs swapped per second\n");
printf("	INTR:	Number of device interrupts per second\n");
printf("	DISK:	Disk traffic in transfers per second\n");
printf("	CNTXT:	Number of context switches per second\n");
printf("	LOAD:	Average number of jobs running over the last minute\n");
printf("	TPKTS:	Total number of incoming/outgoing packets\n");
printf("		on the ethernet per second\n");
printf("	IPKTS:  Incoming packet transfers per second\n");
printf("	OPKTS:	Outgoing packet transfers per second\n");
printf("	COLLS:  Collisions per second detected on the ethernet\n");
printf("	ERRS:	Errors per second on incoming packets\n\n");
		exit(1);

	    case 'p':
		input_option = FROM_PM;
		get_cpulist_from_pm(optarg);
		break;

	    case 'u':
		update = atoi(optarg);
		if(update < 2)
		    update = 2;
		break;

	    case '?':
		printf("\nUsage: cps_sysmon -h [-f <filename>][-p <production_system>][-u <update>]\n");
		exit(1);
	}

    if(input_option == UNDEFINED) {
	input_option = FROM_FILE;
	get_cpulist_from_file("default");
    }

    if(!total_cpus) {
	printf("No cpus found.\n");
	exit(-1);
    }

    establish_remote_connections();
    gethostname(hostnm, sizeof(hostnm));
    truncate(hostnm);

    sigset(SIGINT, sigint_handler);
    initscr();
    nocbreak();
    noecho();

#ifdef IRIX
    attroff(A_BOLD);
#endif
    if(input_option == FROM_PM)
	sprintf(buffer,
		"HOSTNAME:          Update:        PS:                 JM:");
    else
	sprintf(buffer,
		"HOSTNAME:          Update:");
    mvaddstr(0, 0, buffer);

#ifdef IRIX
    attrset(A_BOLD);
#endif
    sprintf(buffer, "%.6s", hostnm);
    mvaddstr(0, 10, buffer);
    sprintf(buffer, "%d sec", update);
    mvaddstr(0, 26, buffer);

    if(input_option == FROM_PM) {
	sprintf(buffer, "%s", p->ps_name);
	mvaddstr(0, 38, buffer);
	sprintf(buffer, "%s", p->jm_cpu->cpu_name);
	mvaddstr(0, 58, buffer);
    }

    sprintf(buffer, "Node    %5s %4s %5s %5s %5s %5s  %5s %5s %5s %5s %5s %5s",
		"CPU", "PAGE", "SWAP", "INTR", "DISK", "CNTXT", "LOAD",
		"TPKTS", "IPKTS", "OPKTS", "COLLS", "ERRS");
    mvaddstr(1, 0, buffer);
    mvaddstr(2, 32, "PLEASE WAIT...");
#ifdef IRIX
    attroff(A_BOLD);
#endif
    refresh();

    get_data();
    get_data();

    while(1)
	sleep(1000);
}


/******************************************************************/
/*
    FUNCTION NAME:

        get_data()

  DESCRIPTION:

        Connects to each node, gets information, and calls display_data()

  FORMAL PARAMETERS:

        None

                                                                  */
/******************************************************************/

void get_data(void)
{
    struct CPULIST *index;
    struct timeval time_out;

    time_out.tv_sec = 10;
    time_out.tv_usec = 0;

    index = head;
    while(index != NULL) {
	index->status = clnt_call(index->client, RSTATPROC_STATS, xdr_void, 0,
		xdr_statstime, &index->stats_info, time_out);
	index = index->next;
    }
    display_data();
    flag = 1;

    signal(SIGALRM, get_data);
    alarm(update);
}


/******************************************************************/
/*
    FUNCTION NAME:

        display_data()

  DESCRIPTION:

        Displays data for each node.
        Converts RSTATD data into displayable data,
            scales it, scrolls the window, and draws it onto the screen(),
            then updates the clock.

  FORMAL PARAMETERS:

        None

                                                                  */
/******************************************************************/

void display_data(void)
{
#define STATS index->stats_info
#define PREV  index->prev_v

    long t;
    int temp, total, cnt = 0, ave = 0;
    char buffer[255];
    struct CPULIST *index;
    struct tm *curt;

    index = head;
    while(index != NULL) {
	if(index->status == RPC_SUCCESS) {
            temp = STATS.cp_time[CP_USER] - PREV.cp_time[CP_USER] +
		   STATS.cp_time[CP_NICE] - PREV.cp_time[CP_NICE] +
		   STATS.cp_time[CP_SYS]  - PREV.cp_time[CP_SYS];
	    total = temp + STATS.cp_time[CP_IDLE] - PREV.cp_time[CP_IDLE];

	    index->data[P_CPU] = (total > 0 ? (temp*100.0/total) : 0.0);

	    ave += index->data[P_CPU];
	    index->data[P_PAGE] = 	STATS.v_pgpgin  - PREV.v_pgpgin +
				      	STATS.v_pgpgout - PREV.v_pgpgout;
	    index->data[P_SWAP] =	STATS.v_pswpin  - PREV.v_pswpin +
				     	STATS.v_pswpout - PREV.v_pswpout;
	    index->data[P_INTR] =	STATS.v_intr - PREV.v_intr;
	    index->data[P_DISK] = 	STATS.dk_xfer[0] - PREV.dk_xfer[0] +
            	    			STATS.dk_xfer[1] - PREV.dk_xfer[1] +
                			STATS.dk_xfer[2] - PREV.dk_xfer[2] +
                			STATS.dk_xfer[3] - PREV.dk_xfer[3];
	    index->data[P_CNTXT] = 	STATS.v_swtch - PREV.v_swtch;
	    index->data[P_LOAD] = 	STATS.avenrun[0];
	    index->data[P_TPKTS] = 	STATS.if_ipackets - PREV.if_ipackets +
					STATS.if_opackets - PREV.if_opackets;
	    index->data[P_IPKTS] = 	STATS.if_ipackets - PREV.if_ipackets;
	    index->data[P_OPKTS] = 	STATS.if_opackets - PREV.if_opackets;
	    index->data[P_COLLS] = 	STATS.if_collisions - PREV.if_collisions;
	    index->data[P_ERRS] = 	STATS.if_ierrors - PREV.if_ierrors +
					STATS.if_oerrors - PREV.if_oerrors;

	    for(temp = 1; temp < MAXITEMS; temp++) {
		if(index->data[temp] < 0)
		    index->data[temp] = 0;
		if(temp != P_LOAD && temp != P_CPU)
		    index->data[temp] /= update;
	    }

	    if(flag) {
	        sprintf(buffer,
		"%8.8s:%4d %4d %5d %5d %5d %5d %6.3f %5d %5d %5d %5d %5d ",
		index->name, index->data[P_CPU], index->data[P_PAGE],
		index->data[P_SWAP], index->data[P_INTR], index->data[P_DISK],
		index->data[P_CNTXT], (float)(index->data[P_LOAD]/256.0),
		index->data[P_TPKTS], index->data[P_IPKTS],
		index->data[P_OPKTS], index->data[P_COLLS],
		index->data[P_ERRS]);
		mvaddstr(cnt+2, 0, buffer);
		cnt++;
	    }
	} else {
	    sprintf(buffer, "%8.8s:%70s ", index->name,
"                          Not Available                                 ");
	    mvaddstr(cnt+2, 0, buffer);
	    cnt++;
	}
	index->prev_v = index->stats_info;
	index = index->next;
    }
    if(flag) {
	sprintf(buffer, " AVG CPU:%4d ", (int)(ave/cnt));
	mvaddstr(cnt+2, 0, buffer);
	t = time((long *)0);
	curt = localtime(&t);
	sprintf(buffer, "%.2d:%.2d:%.2d", curt->tm_hour, curt->tm_min, curt->tm_sec);
	mvaddstr(0, 68, buffer);
	refresh();
    }
}


/******************************************************************/
/*
    FUNCTION NAME:

        void get_cpulist_from_file(char *file)

  DESCRIPTION:

        Reads a list of node name from a file and stores it in the
        CPULIST.

  FORMAL PARAMETERS:

        char *file:	Name of the file to load the cpu list from.

                                                                  */
/******************************************************************/

void get_cpulist_from_file(char *file)
{
    FILE *fp;
    char filename[81], name[81];
    struct CPULIST *index, *pointers[300];
    int i, n = 0;

/* Get the users home path */

    if(!strcmp(file, "default"))
        sprintf(filename, "%s/%s", getenv("HOME"), ".sysmon");
    else
        strcpy(filename, file);

/* Try to open the file for reading */

    if((fp = fopen(filename, "r")) == NULL) {
	fprintf(stderr, "Cannot locate the %s file.\n\n", filename);
	exit(-1);
    }

/* Read each name and add it to the CPULIST */

    while(fscanf(fp, "%s", name) != EOF && total_cpus <= num_display_lines)
	acpu(name);

    fclose(fp);

/* Create an array of pointers to each element so that they may be sorted */

    index = head;
    while(index != NULL) {
        pointers[n] = index;
        n++;
        index = index->next;
    }
    pointers[n] = NULL;

/* Sort the array */

    qsort(pointers, n, sizeof(struct CPULIST *), compare);

/* Re-link the list */

    head = pointers[0];
    for(i = 0; i < n; i++)
        pointers[i]->next = pointers[i+1];
}


/******************************************************************/
/*
    FUNCTION NAME:

        void get_cpulist_from_pm(void)

  DESCRIPTION:

        Get the list of nodes which belong to the Production System
        from the Job Manager.

  FORMAL PARAMETERS:

        name - The name of the Production System

                                                                  */
/******************************************************************/

void get_cpulist_from_pm(char *name)
{
    struct CPULIST *index, *pointers[300];
    struct prod_cpu_list *pc;
    int i, n;

/* Connect to the PM */

    if(pm_connect() == -1) {
	printerr(FACILITY, FATAL_MSG, "Failed to connect to a PM", NULL, 1);
	exit(0);
    }

/* Ask for a PS by name */

    p = find_prodsys(db.ps, name);
    if(p == NULL) {
	printerr(FACILITY, FATAL_MSG, "Production system not found", NULL, 2);
	exit(0);
    }

/* Add the JM node first */

    acpu(p->jm_cpu->cpu_name);

/* Add the rest, but make sure they're ot the JM's node */

    pc = p->cpus;
    while(pc != NULL && total_cpus <= num_display_lines) {
        if(strcmp(pc->cpu->cpu_name, p->jm_cpu->cpu_name))
	    acpu(pc->cpu->cpu_name);
	pc = pc->next;
    }

/* Begin the sort routine */
/* A separate set of sorting routines is required for
   get_cpulist_from_pm and get_cpulist_from_file because
   here, we want the JM node to always be first */

/* Build an array of pointers */

    n = 0;
    index = head->next;
    while(index != NULL) {
        pointers[n] = index;
        n++;
        index = index->next;
    }
    pointers[n] = NULL;

/* Sort the node names */

    qsort(pointers, n, sizeof(struct CPULIST *), compare);

/* Re-link the list */

    head->next = pointers[0];
    for(i = 0; i < n; i++)
        pointers[i]->next = pointers[i+1];
}

static int compare(struct CPULIST **x, struct CPULIST **y)
{
    return(strcmp((*x)->name, (*y)->name));
}


/******************************************************************/
/*
    FUNCTION NAME:

        void acpu(char *name)

  DESCRIPTION:

        Create a link in the CPULIST list and add the node to it.
        Also setup preliminary networking stuff.

  FORMAL PARAMETERS:

        name - The name of the node

                                                                  */
/******************************************************************/

void acpu(char *name)
{
    struct hostent *hp;
    struct CPULIST *new, *index;

/* Get information about the remote nodes */

    if((hp = gethostbyname(name)) == NULL) {
	fprintf(stderr, "Host not found: %.6s\n", name);
	return;
    }

/* Create a new link */

    new = (struct CPULIST *)malloc(sizeof(struct CPULIST));

/* Determine where to add it */

    if(head == NULL)
	head = index = new;
    else {
	index = head;
	while(index->next != NULL) 
	    index = index->next;
	index->next = new;
	index = new;
    }
    total_cpus++;

/* Truncate the name after the first '.' */

    truncate(name);

/* Copy it to the CPULIST list */

    strcpy(index->name, name);

/* Setup some RPC necessities */

    bcopy(hp->h_addr, (caddr_t)&index->server_addr.sin_addr, hp->h_length);
    index->server_addr.sin_family = AF_INET;
    index->server_addr.sin_port = 0;
    index->status = RPC_SUCCESS;
    index->sock = RPC_ANYSOCK;
    index->next = NULL;
}


/******************************************************************/
/*
    FUNCTION NAME:

        void rcpu(char *name)

  DESCRIPTION:

        Parse the CPULIST list and remove 'name' from the list.

  FORMAL PARAMETERS:

        name - The name of the node

                                                                  */
/******************************************************************/

void rcpu(char *name)
{
    struct CPULIST *index;

    index = head;
    if(!strcmp(index->name, name)) {
	free(index);
	head = index = index->next;
    }
    else {
	while(strcmp(index->next->name, name))
	    index = index->next;
	free(index->next);
	index->next = index->next->next;
    }
}


/******************************************************************/
/*
    FUNCTION NAME:

        void truncate(char *name)

  DESCRIPTION:

        Truncate an Internet name after the first '.'

        EX: fnsfa.fnal.gov --> fnsfa

  FORMAL PARAMETERS:

        name - The name of the node

                                                                  */
/******************************************************************/

void truncate(char *p)
{
    register int n;

/* Brute force, but necessary */

    for(n = 0; n <= strlen(p); n++) {
	if(p[n] == '.')
	    break;
    }
    p[n] = '\0';
}


/******************************************************************/
/*
    FUNCTION NAME:

        void destroy_list(void)

  DESCRIPTION:

        Destroys all links in the CPULIST, and frees all allocated memory

  FORMAL PARAMETERS:

        None

                                                                  */
/******************************************************************/

void destroy_list(void)
{
    struct CPULIST *index;

    index = head;
    while(index != NULL) {
	free(index);
	index = index->next;
    }
}


/******************************************************************/
/*
    FUNCTION NAME:

        void establish_remote_connections(void)

  DESCRIPTION:

        Establishes RPC remote connections with each node
        in the CPULIST list.

  FORMAL PARAMETERS:

        None

                                                                  */
/******************************************************************/

void establish_remote_connections(void)
{
    struct CPULIST *index;
    struct timeval time_out;

/* Number of seconds RPC will WAIT before attempting another connection.
   NOT the number of seconds before it FAILs a connection */

    time_out.tv_sec = 5;
    time_out.tv_usec = 0;

/* Create UDP connections with the RSTATD daemon running on the remote nodes */

    index = head;
    while(index != NULL) {
	printf("\nAttempting connection with %s...", index->name); fflush(stdout);
	if((index->client = clntudp_create(&index->server_addr, RSTATPROG,
		RSTATVERS_TIME, time_out, &index->sock)) == NULL) {
	    fprintf(stderr,
		"\nSYSMON: Cannot establish connection%s",
		clnt_spcreateerror(""));

/* Failed connection, remove node from the list */

	    rcpu(index->name);
            total_cpus--;
	}
	else {

/* Established connections, authenticate it */

	    index->client->cl_auth = authunix_create_default();
	}
	index = index->next;
    }
}


/******************************************************************/
/*
    FUNCTION NAME:

        void sigint_handler(int sig, int code, struct sigcontext *sc)

  DESCRIPTION:

        Handle all interrupt signals so that we can exit gracefully.
        Close all socket connections, free all memory allocated to the
        pixmaps, graphic contexts, and CPULIST.

  FORMAL PARAMETERS:

        sig     - Caught signal
        code    - Unused for this signal
        sc      - The processors context

                                                                  */
/******************************************************************/

void sigint_handler(int sig, int code, struct sigcontext *sc)
{
    struct CPULIST *index;

    move(LINES-1, 0);
    clrtoeol();
    refresh();
    endwin();

    index = head;
    while(index != NULL) {
	clnt_destroy(index->client);
	close(index->sock);
	free(index);
	index = index->next;
    }
    exit(0);
}

display_error(msg)
char *msg;
{
    printf("%s\n",msg);
    fflush(stdout);
}
