/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: db.c

    Explanation: Database manipulation routines.

*/

/*****************/
/* Include Files */
/*****************/

#include <stdio.h>
#ifndef R3000
#include <stdlib.h>
#endif
#include "../inc/db_param.h"
#include "../inc/db_struct.h"
#include "../inc/db_externs.h"
#include "../inc/perr.h"

#define FACILITY "DBLIB"

extern char *msg_cmd, **msg_args;	/* Defined in MSGFMT.C */
extern FILE *fopen();
extern char *fgets();
extern char *ctime();

/******************************************************************/
/*
    FUNCTION NAME:

      find_fs

  DESCRIPTION:

      Returns a pointer to the filesystem with a given internal id.

  FORMAL PARAMETERS:

      struct fs_list *fl - Filesystem list to search.

      int id - Filesystem id.

  FUNCTION VALUE:

      (struct fs_list *) - Pointer to the filesystem or NULL.

								  */
/******************************************************************/

struct fs_list *find_fs(fl,id)
struct fs_list *fl;
int id;
{
	while ((fl != NULL) && (fl->id != id)) fl = fl->next;
	return(fl);
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_fs

  DESCRIPTION:

      Adds a filesystem to a filesystem list.

  FORMAL PARAMETERS:

      struct fs_list **fl - First filesystem in the list.

      struct cpu_list *c - Pointer to the CPU where the disk is located.

      char *mount_point	- Mount point on the CPU the disk is at.

      int blocks - Number of blocks on the disk.

      int id - Internal id of the filesystem.


								  */
/******************************************************************/

add_fs(fl,c,mount_point,blocks,id)
struct fs_list **fl;
struct cpu_list *c;
char *mount_point;
int blocks;
int id;
{
	char buf[80];

	struct fs_list *f;

	f = find_fs(*fl,id);    /* Check to see if the filesystem is already
				       defined */
	if (f != NULL) return;

	f = (struct fs_list *) hmalloc(sizeof(struct fs_list));
	f->id = id;
	f->mark = 0;
	f->refcount = 0;
	f->blocks = blocks;

	f->cpu = c;
	c->refcount++;  /* Increment reference count of CPU */

	strcpy(buf,mount_point);
	f->mount_point = (char *) hmalloc(strlen(buf)+1);
	strcpy(f->mount_point,buf);

	f->next = *fl;
	*fl = f;
}

/******************************************************************/
/*
    FUNCTION NAME:

      remove_fs

  DESCRIPTION:

      Removes a filesystem from a list of filesystems

  FORMAL PARAMETERS:

      struct fs_list **fs - List of filesystems to search.

      int id - Internal id of the filesystem.

                                                                  */
/******************************************************************/

int remove_fs(fl,id)
struct fs_list **fl;
int id;
{

	struct fs_list *f,*f1;

	f = find_fs(*fl,id);
	if (f == NULL) return(0);  /* Filesystem not found */

	if (f->refcount > 0) return(-1);	/* Other structures still reference 
						   this record */

	if (*fl == f) {	    /* Remove first in list */
		*fl = f->next;
	}
	else {
		f1 = *fl;
		while (f1->next != f) f1 = f1->next;
		f1->next = f->next;
	}

	hfree(f->mount_point);
	f->cpu->refcount--;
	hfree(f);

	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_cpu

  DESCRIPTION:

      Searches a CPU list and returns a pointer to the CPU record with
      the given name.

  FORMAL PARAMETERS:

      struct cpu_list *cpu - First CPU in the list.
      char *cpu_name - Name of the CPU.

  FUNCTION VALUE:

      struct cpu_list * - Returns a pointer to the CPU or NULL if not found.

								  */
/******************************************************************/

int cpu_name_compare(cpu1,cpu2)
char *cpu1,*cpu2;
{
	char buf1[80],buf2[80],*b1,*b2,*strchr();

	strcpy(buf1,cpu1);
	b1 = strchr(buf1,'.');
	if (b1 != NULL) *b1 = '\0';

	strcpy(buf2,cpu2);
	b2 = strchr(buf2,'.');
	if (b2 != NULL) *b2 = '\0';

	return(strcmp(buf1,buf2));
}

struct cpu_list *find_cpu(cpu,cpu_name)
struct cpu_list *cpu;
char *cpu_name;
{
	struct cpu_list *c;

	c = cpu;
	while (c != NULL) {
		if (cpu_name_compare(cpu_name,c->cpu_name) == 0) return(c);
		c = c->next;
	}
	return(c);
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_tape

  DESCRIPTION:

      Searches a tape list and returns a pointer to the tape record with
      the given device name & cpu name

  FORMAL PARAMETERS:

      struct tape_drive_list *tape - First record in the tape list.
      char *dev_name - Name of the device.
      char *cpu_name - CPU the device is on.

  FUNCTION VALUE:

      struct tape_drive_list * - Returns a pointer to the tape 
				 or NULL if not found.

								  */
/******************************************************************/

struct tape_drive_list *find_tape(tape,dev_name,cpu_name)
struct tape_drive_list *tape;
char *dev_name;
char *cpu_name;
{
	struct tape_drive_list *t;

	t = tape;
	while (t != NULL) {
		if (strcmp(t->dev_name,dev_name) == 0) {
			if (cpu_name_compare(t->cpu->cpu_name,cpu_name) == 0) return(t);
		}
		t = t->next;
	}
	return(t);
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_phys_tape

  DESCRIPTION:

      Searches a tape list and returns a pointer to the tape record with
      the given physical device name & cpu name

  FORMAL PARAMETERS:

      struct tape_drive_list *tape - First record in the tape list.
      char *dev_name - Physical name of the device.
      char *cpu_name - CPU the device is on.

  FUNCTION VALUE:

      struct tape_drive_list * - Returns a pointer to the tape 
				 or NULL if not found.

								  */
/******************************************************************/

struct tape_drive_list *find_phys_tape(tape,dev_name,cpu_name)
struct tape_drive_list *tape;
char *dev_name;
char *cpu_name;
{
	struct tape_drive_list *t;

	t = tape;
	while (t != NULL) {
		if (strcmp(t->physical,dev_name) == 0) {
			if (cpu_name_compare(t->cpu->cpu_name,cpu_name) == 0) return(t);
		}
		t = t->next;
	}
	return(t);
}

/************************************************************************/
/*

    Function: sort_tape_drives

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int t_compare(x,y)
struct tape_drive_list **x,**y;
{
	return(strcmp((*x)->dev_name,(*y)->dev_name));
}

int sort_tape_drives(tlist)
struct tape_drive_list **tlist;
{
	struct tape_drive_list *t,*td[300];

	int i,n;

	/* Set up array of pointers */

	n = 0;
	t = *tlist;
	while (t != NULL) {
		if (n >= 300) {
			printerr(FACILITY,ERROR_MSG,"Too many tape drives",NULL,7);
			return;
		}
		td[n] = t;
		n++;
		t = t->next;
	}
	td[n] = NULL;

	/* Sort the tapes */

	qsort(td,n,4,t_compare);

	/* Re-link the list */

	*tlist = td[0];
	for (i=0; i<n; i++) td[i]->next = td[i+1];

	return;
}

/******************************************************************/
/*
    FUNCTION NAME:

      cpu_alias

  DESCRIPTION:

      Takes a full CPU name and returns its alias.

  FORMAL PARAMETERS:

      char *name - Full CPU name.

  FUNCTION VALUE:

      (char *) - CPU name with domain information stripped off.

								  */
/******************************************************************/

char *cpu_alias(name)
char *name;
{
	static char alias[64];
	char *c,*strchr();

	strcpy(alias,name);
	c = strchr(alias,'.');
	if (c != NULL) *c = 0;
	return(alias);
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_cpu_alias

  DESCRIPTION:

      Searches a CPU list and returns a pointer to the CPU record with
      the given alias.

  FORMAL PARAMETERS:

      struct cpu_list *cpu - First CPU in the list.
      char *cpu_name - Alias of the CPU.

  FUNCTION VALUE:

      struct cpu_list * - Returns a pointer to the CPU or NULL if not found.

								  */
/******************************************************************/

struct cpu_list *find_cpu_alias(cpu,alias)
struct cpu_list *cpu;
char *alias;
{
	struct cpu_list *c;

	c = cpu;
	while (c != NULL) {
		if (cpu_name_compare(cpu_alias(c->cpu_name),alias) == 0) return(c);
		c = c->next;
	}
	return(c);
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_cpu

  DESCRIPTION:

      Adds a CPU to a CPU list

  FORMAL PARAMETERS:

      struct cpu_list **cpu - First CPU in the list.
      char *cpu_name - Name of the CPU.
      int cpu_type - Type of CPU.
      int power - Power of the CPU where 100 = 1 VAX 11/780.

  FUNCTION VALUE:

      struct cpu_list * - Returns a pointer to the new CPU record.  If the
			  CPU was already there, nothing is added and 
			  a pointer to the CPU is returned.

								  */
/******************************************************************/

struct cpu_list *add_cpu(cpu,cpu_name,cpu_type,power)
struct cpu_list **cpu;
char *cpu_name;
int cpu_type,power;
{

	struct cpu_list *c;

	c = find_cpu(*cpu,cpu_name);
	if (c != NULL) {
		return(c);	/* CPU already exists */
	}

	c = (struct cpu_list *) hmalloc(sizeof(struct cpu_list));
	c->cpu_name = (char *) hmalloc(strlen(cpu_name)+1);
	strcpy(c->cpu_name,cpu_name);
	c->cpu_type = cpu_type;
	c->opsys = DB_UNIX;
	c->refcount = 0;
	c->power = power;
	c->np = 0;
	c->status = DB_WORKING;
	c->next = *cpu;
	c->task = DB_TASK_ANY;
	*cpu = c;

	return(c);
}

/******************************************************************/
/*
    FUNCTION NAME:

      delete_cpu

  DESCRIPTION:

      Removes a CPU from a CPU list.

  FORMAL PARAMETERS:

      struct cpu_list **cpu - First CPU in the list.
      char *cpu_name - Name of the CPU.

  FUNCTION VALUE:

      int - Returns 0 if the CPU was removed successfully.  Returns -1 if
	    the CPU could not be removed because of a refcount > 0.

								  */
/******************************************************************/

int delete_cpu(cpu,cpu_name)
struct cpu_list **cpu;
char *cpu_name;
{

	struct cpu_list *c,*pc;

	c = find_cpu(*cpu,cpu_name);
	if (c != NULL) {

		if (c->refcount > 0) return(-1);    /* Record is still referenced */

		if (*cpu == c) {
			*cpu = c->next;
		}
		else {
			pc = *cpu;
			while(pc->next != c) pc = pc->next;
			pc->next = c->next;
		}

		hfree(c->cpu_name);
		hfree(c);
	}

	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_tape

  DESCRIPTION:

      Adds a tape to a tape drive list.

  FORMAL PARAMETERS:

      struct tape_drive_list **tape - First record in the list.
      char *dev_name - Name of the device to add.
      char *physical - Physical name of the device.
      int dev_type - Type of device to add.
      struct cpu_list *cpu - CPU from which the tape can be accessed.

  FUNCTION VALUE:
 
      struct tape_drive_list * - Returns a pointer to the new tape record.
	  If the tape already exists, then the pointer to the record is 
	  returned.

								  */
/******************************************************************/

struct tape_drive_list *add_tape(tape,dev_name,physical,dev_type,cpu)
struct tape_drive_list **tape;
char *dev_name,*physical;
struct cpu_list *cpu;
int dev_type;
{
	struct tape_drive_list *t;

	t = find_tape(*tape,dev_name,cpu->cpu_name);
	if (t != NULL) return(t);	/* Tape already exists */

	cpu->refcount++;

	t = (struct tape_drive_list *) hmalloc(sizeof(struct tape_drive_list));
	t->dev_name = (char *) hmalloc(strlen(dev_name)+1);
	strcpy(t->dev_name,dev_name);
	t->physical = (char *) hmalloc(strlen(physical)+1);
	strcpy(t->physical,physical);
	t->dev_type = dev_type;
	t->status = DB_WORKING;
	t->cpu = cpu;
	t->used = DB_UNALLOC;
	t->refcount = 0;
	t->next = *tape;
	*tape = t;

	t->mi = (struct mount_info *) hmalloc(sizeof(struct mount_info));
	t->mi->vsn[0] = '\0';
	t->mi->rstate = 0;
	t->mi->mstate = DB_NOT_MOUNTED;

	return(t);
}

/******************************************************************/
/*
    FUNCTION NAME:

      delete_tape

  DESCRIPTION:

      Removes a tape from a tape list.

  FORMAL PARAMETERS:

      struct tape_drive_list **tape - First record in the list.
      char *dev_name - Name of the device to add.
      char *cpu_name - Name of the CPU the device can be read from.

  FUNCTION VALUE:
 
      int - Returns 0 if the tape was removed successfully.  Returns -1 if
	    the tape could not be removed because of a refcount > 0.

								  */
/******************************************************************/

int delete_tape(tape,dev_name,cpu_name)
struct tape_drive_list **tape;
char *dev_name,*cpu_name;
{
	struct tape_drive_list *t,*pt;

	t = find_tape(*tape,dev_name,cpu_name);
	if (t != NULL) {

		if (t->refcount > 0) return(-1);

		t->cpu->refcount--;

		if (*tape == t) {
			*tape = t->next;
		}
		else {
			pt = *tape;
			while(pt->next != t) pt = pt->next;
			pt->next = t->next;
		}

		hfree(t->dev_name);
		hfree(t->physical);
		hfree(t->mi);
		hfree(t);
	}

	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_prodsys

  DESCRIPTION:

      Searches a list of production systems for the one specified.

  FORMAL PARAMETERS:

      struct prod_sys *ps - First production system in the list.
      char *prd_name - Name of the production system to search for.

  FUNCTION VALUE:

      struct prod_sys * - Returns a pointer to the production system record.
	    Returns NULL if the production system wasn't found.
								  */
/******************************************************************/

struct prod_sys *find_prodsys(ps,prd_name)
struct prod_sys *ps;
char *prd_name;
{
	struct prod_sys *p;

	p = ps;
	while (p != NULL) {
		if (strcmp(p->ps_name,prd_name) == 0) return(p);
		p = p->next;
	}
	return(p);
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_prodsys

  DESCRIPTION:

      Adds a production system to a list.

  FORMAL PARAMETERS:

      struct prod_sys **ps - First production system in the list.
      char *prd_name - Name of the system to add.
      struct cpu_list *jm_cpu - CPU that will run the JM.

  FUNCTION VALUE:

      struct prod_sys * - Returns a pointer to the production system added.
	    If the system already exits, it is not added again and a pointer
	    to the system is returned.
								  */
/******************************************************************/

struct prod_sys *add_prodsys(ps,prd_name,jm_cpu)
struct prod_sys **ps;
char *prd_name;
struct cpu_list *jm_cpu;
{
	struct prod_sys *p;

	p = find_prodsys(*ps,prd_name);
	if (p != NULL) {
		return(p);	/* Production system already exists */
	}

	jm_cpu->refcount++;
	jm_cpu->np++;

	p = (struct prod_sys *) hmalloc(sizeof(struct prod_sys));
	p->ps_name = (char *) hmalloc(strlen(prd_name)+1);
	strcpy(p->ps_name,prd_name);
	p->jm_cpu = jm_cpu;
	p->jm_fs = NULL;
	p->account = NULL;
	p->passwd = NULL;
	p->login_dir = NULL;
	p->cpus = NULL;
	p->tapes = NULL;
	p->next_entry = 1;
	p->next_area = 1;
	p->job_area[0] = 0;
	p->job_area[1] = 0;
	p->job_area[2] = 0;
	p->newjob = NULL;
	p->back = NULL;
	p->job_queue = NULL;
	p->on_deck = NULL;
	p->current = NULL;
	p->rundown = NULL;
	p->status = jm_cpu->status;
	p->users = NULL;
	p->next = *ps;
	*ps = p;

	return(p);
}

/************************************************************************/
/*

    Function:  authorized_user

    Description:

    Arguments:

    Function Value:  -1 if the username is authorized for a production
		     system, otherwize 0.

*/
/************************************************************************/

int authorized_user(p,username)
struct prod_sys *p;
char *username;
{
	struct word_list *u;

	u = p->users;
	while (u != NULL) {
		if (strcmp(u->word,"*") == 0) return -1;
		if (strcmp(u->word,username) == 0) return -1;
		u = u->next;
	}
	return 0;
}

/************************************************************************/
/*

    Function: select_area

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int select_area(p)
struct prod_sys *p;
{
	int area;
	int counter = 0;

	area = p->next_area;
	if (area < 1) area = 1;
	if (area > 3) area = 3;

try_again:
	if (p->job_area[area-1] == -1) {
		if (counter > 3) return -1;
		counter++;
		area++;
		if (area > 3) area = 1;
		goto try_again;
	}
	p->next_area++;
	if (p->next_area > 3) p->next_area = 1;
	return(area);
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_prod_cpu_list

  DESCRIPTION:

      Deallocates a production system's list of CPUs.  CPU reference counts
      are adjusted.

  FORMAL PARAMETERS:

      struct prod_cpu_list *pcl;

								  */
/******************************************************************/

clear_prod_cpu_list(pcl)
struct prod_cpu_list *pcl;
{

	struct prod_cpu_list *pcl1;

	while (pcl != NULL) {
		pcl1 = pcl->next;
		pcl->cpu->refcount--;		    /* Decrement the CPUs refcount */
		pcl->cpu->np -= pcl->max_proc;
		pcl->fs->refcount--;
		hfree(pcl);
		pcl = pcl1;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_prod_tape_list

  DESCRIPTION:

      Deallocates a production system's list of tapes.  Tape reference counts
      are adjusted.

  FORMAL PARAMETERS:

      struct prod_tape_list *ptl;

								  */
/******************************************************************/

clear_prod_tape_list(ptl)
struct prod_tape_list *ptl;
{
	struct prod_tape_list *ptl1;

	while(ptl != NULL) {
		ptl1 = ptl->next;
		ptl->tape->refcount--;		    /* Decrement the tape's refcount */
		hfree(ptl);
		ptl = ptl1;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      delete_prodsys

  DESCRIPTION:

      Removes a production system from a list of systems.

  FORMAL PARAMETERS:

      struct prod_sys **ps - First system in the list.
      char *prd_name - Name of the production system to remove.

  FUNCTION VALUE:

      none
								  */
/******************************************************************/

delete_prodsys(ps,prd_name)
struct prod_sys **ps;
char *prd_name;
{
	struct cpu_list *c;
	struct prod_sys *p,*pp;

	struct prod_cpu_list *pc,*npc;
	struct prod_tape_list *pt,*npt;

	p = find_prodsys(*ps,prd_name);
	if (p != NULL) {

		/* Remove CPUs from production system */

		clear_prod_cpu_list(p->cpus);

		/* Remove tapes from production system */

		clear_prod_tape_list(p->tapes);

		/* Remove jobs from the production system */

		clear_prod_job_list(p->job_queue);
		clear_prod_job_list(p->on_deck);
		clear_prod_job_list(p->current);
		clear_prod_job_list(p->rundown);

		/* Remove authorized users */

		clear_word_list(p->users);

		/* Remove production system itself */

		if (*ps == p) {
			*ps = p->next;
		}
		else {
			pp = *ps;
			while(pp->next != p) pp = pp->next;
			pp->next = p->next;
		}

		p->jm_cpu->refcount--;
		p->jm_cpu->np--;

		p->jm_fs->refcount--;

		if (p->ps_name != NULL) hfree(p->ps_name);
		if (p->account != NULL) hfree(p->account);
		if (p->passwd != NULL) hfree(p->passwd);
		if (p->login_dir != NULL) hfree(p->login_dir);
		hfree(p);
	}

	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_cpu_in_prod

  DESCRIPTION:

      Finds a CPU in a production system.

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

struct prod_cpu_list *find_cpu_in_prod(cpu,p,cpu_name)
struct cpu_list *cpu;
struct prod_sys *p;
char *cpu_name;
{
	struct cpu_list *c;
	struct prod_cpu_list *pcl;

	/* Get a pointer to the CPU */

	c = find_cpu(cpu,cpu_name);
	if (c == NULL) return(NULL);

	/* Find the same pointer in the production system's CPU list */

	pcl = p->cpus;
	while (pcl != NULL) {
		if (pcl->cpu == c)
			return(pcl);
		else
			pcl = pcl->next;
	}
	return(NULL);

}

/******************************************************************/
/*
    FUNCTION NAME:

      add_cpu_to_prod

  DESCRIPTION:

      Adds a CPU to a production system.

  FORMAL PARAMETERS:

      struct prod_sys *ps - Production system to modify.
      struct cpu_list *cpu - CPU to add.
      struct fs_list *fs - Filesystem that will be used.
      int maxproc - Maximum number of processes allowed to run.
      char *mp - Mount point.

  FUNCTION RETURN:

      none
								  */
/******************************************************************/

add_cpu_to_prod(ps,cpu,fs,maxproc,mp,lognum)
struct prod_sys *ps;
struct cpu_list *cpu;
struct fs_list *fs;
int maxproc,lognum;
char *mp;
{
	struct prod_cpu_list *pcl;

	cpu->refcount++;
	cpu->np += maxproc;
	fs->refcount++;

	pcl = find_cpu_in_prod(cpu,ps,cpu->cpu_name);
	if (pcl == NULL) {
		pcl = (struct prod_cpu_list *) hmalloc(sizeof(struct prod_cpu_list));
		pcl->next = ps->cpus;
		ps->cpus = pcl;
	}

	pcl->cpu = cpu;
	pcl->max_proc = maxproc;
	pcl->fs = fs;
	pcl->lognum = lognum;
}

/******************************************************************/
/*
    FUNCTION NAME:

      delete_cpu_from_prod

  DESCRIPTION:

      Removes a CPU from a production system.

  FORMAL PARAMETERS:

      struct prod_sys *ps - Production system to modify.
      struct cpu_list *cpu - CPU to delete from the system.

  FUNCTION RETURN:

      none
								  */
/******************************************************************/

delete_cpu_from_prod(ps,cpu)
struct prod_sys *ps;
struct cpu_list *cpu;
{
	struct prod_cpu_list *pc,*ppc;
	struct fs_list *f;

	/* Find the record to delete */

	pc = ps->cpus;
	while (pc->cpu != cpu) {
		pc = pc->next;
		if (pc == NULL) return;
	}

	/* Delete the record */

	if (ps->cpus == pc) {
		ps->cpus = pc->next;
	}
	else {
		ppc = ps->cpus;
		while(ppc->next != pc) ppc = ppc->next;
		ppc->next = pc->next;
	}

	/* Update the CPU record */

	cpu->refcount--;
	cpu->np -= pc->max_proc;

	/* Update the filesystem record */

	pc->fs->refcount--;

	hfree(pc);

}

/******************************************************************/
/*
    FUNCTION NAME:

      add_tape_to_prod

  DESCRIPTION:

      Adds a tape to a production system.

  FORMAL PARAMETERS:

      struct prod_sys *ps - System to modify.
      struct tape_drive_list *tape - Tape to add.

  FUNCTION VALUE:

      int - 0 if successful, -1 if the tape is already part of another
	production system.
								  */
/******************************************************************/

int add_tape_to_prod(ps,tape)
struct prod_sys *ps;
struct tape_drive_list *tape;
{
	struct prod_tape_list *pt;

	if (tape->refcount > 0) return(-1);	/* Already part of another prodsys */

	tape->refcount++;

	pt = (struct prod_tape_list *) hmalloc(sizeof(struct prod_tape_list));
	pt->tape = tape;
	pt->next = ps->tapes;
	ps->tapes = pt;

	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      delete_tape_from_prod

  DESCRIPTION:

      Removes a tape from a production system.

  FORMAL PARAMETERS:

      struct prod_sys *ps - Production system to modify.
      struct tape_drive_list *tape - Tape to delete.

								  */
/******************************************************************/

delete_tape_from_prod(ps,tape)
struct prod_sys *ps;
struct tape_drive_list *tape;
{
	struct prod_tape_list *pt,*ppt;

	/* Find the record to delete from the list */

	pt = ps->tapes;
	while(pt->tape != tape) {
		pt = pt->next;
		if (pt == NULL) return;
	}

	/* Delete it from the list */

	if (ps->tapes == pt) {
		ps->tapes = pt->next;
	}
	else {
		ppt = ps->tapes;
		while(ppt->next != pt) ppt = ppt->next;
		ppt->next = pt->next;
	}

	/* Update the tape record */

	tape->refcount--;
	hfree(pt);
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_production_system

  DESCRIPTION:

      Clears an entire list off production systems.

  FORMAL PARAMETERS:

      struct prod_sys *ps - First system in the list.

  FUNCTION VALUE:

      struct prod_sys * - NULL value is returned.
								  */
/******************************************************************/

struct prod_sys *clear_production_systems(ps)
struct prod_sys *ps;
{

	struct prod_sys *p;

	while (ps != NULL) {
		p = ps->next;
		clear_prod_job_list(ps->job_queue);
		clear_prod_job_list(ps->on_deck);
		clear_prod_job_list(ps->current);
		clear_prod_job_list(ps->rundown);
		clear_word_list(ps->users);
		clear_prod_tape_list(ps->tapes);
		clear_prod_cpu_list(ps->cpus);
		ps->jm_cpu->refcount--;		/* Decrement JM CPU's refcount */
		ps->jm_cpu->np--;
		hfree(ps->ps_name);
		hfree(ps);
		ps = p;
	}
	return(ps);
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_fs_list

  DESCRIPTION:

      Deallocates an entire filesystem list.

  FORMAL PARAMETERS:

      struct fs_list *fl - Pointer to the first filesystem in the list.

                                                                  */
/******************************************************************/

struct fs_list *clear_fs_list(fl)
struct fs_list *fl;
{
	while (fl != NULL) {
		hfree(fl->mount_point);
		fl->cpu->refcount--;
		hfree(fl);
		fl = fl->next;
	}
	return(fl);
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_cpu_list

  DESCRIPTION:

      Deallocates an entire CPU list.

  FORMAL PARAMETERS:

      struct cpu_list *c - First record in the list.

  FUNCTION VALUE:

      struct cpu_list * -  NULL value if successful.  Non-NULL if one or
	    more CPUs had non-zero reference counts.

								  */
/******************************************************************/

struct cpu_list *clear_cpu_list(c)
struct cpu_list *c;
{
	struct cpu_list *c1;

	c1 = c;
	while (c1 != NULL) {
		if (c1->refcount != 0) return(c);
		c1 = c1->next;
	}

	while (c != NULL) {
		c1 = c->next;
		hfree(c->cpu_name);
		hfree(c);
		c = c1;
	}

	return(c);
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_tape_list

  DESCRIPTION:

      Deallocates an entire tape drive list.

  FORMAL PARAMETERS:

      struct tape_drive_list *t - First record in the list.

  FUNCTION VALUE:

      struct tape_drive_list * - NULL if successful.  Non-NULL if one or more
	    tapes had non-zero reference counts.

								  */
/******************************************************************/

struct tape_drive_list *clear_tape_list(t)
struct tape_drive_list *t;
{
	struct tape_drive_list *t1;

	t1 = t;
	while (t1 != NULL) {
		if (t1->refcount != 0) return(t);
		t1 = t1->next;
	}

	while (t != NULL) {
		t1 = t->next;
		t->cpu->refcount--;	    /* Decrement the CPU's refcount */
		hfree(t->dev_name);
		hfree(t->physical);
		hfree(t->mi);
		hfree(t);
		t = t1;
	}

	return(t);
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_database

  DESCRIPTION:

      Wipes out the entire database.

  FORMAL PARAMETERS:

      struct fs_list **fs - List of filesystems.

      struct cpu_list **cpu - List of CPUs.

      struct tape_drive_list **tape - List of tapes.

      struct prod_sys *ps - List of production systems.

								  */
/******************************************************************/

clear_database(fs,cpu,tape,ps)
struct fs_list **fs;
struct cpu_list **cpu;
struct tape_drive_list **tape;
struct prod_sys **ps;
{
	*fs = clear_fs_list(*fs);
	*ps = clear_production_systems(*ps);
	*tape = clear_tape_list(*tape);
	*cpu = clear_cpu_list(*cpu);
}

/******************************************************************/
/*
    FUNCTION NAME:

      cpu_in_prodsys

  DESCRIPTION:

      Determines whether a CPU is in a production system's list of CPUs.

  FORMAL PARAMETERS:

      struct prod_sys *ps - Production system to search.

      char *cpu_name - CPU to look for.

  FUNCTION VALUE:

      struct prod_cpu_list * - NULL if the CPU was not found.  Pointer to
		the list record if it was found.
								  */
/******************************************************************/

struct prod_cpu_list *cpu_in_prodsys(ps,cpu_name)
struct prod_sys *ps;
char *cpu_name;
{
	struct prod_cpu_list *pcl;

	pcl = ps->cpus;
	while (pcl != NULL) {
		if (cpu_name_compare(pcl->cpu->cpu_name,cpu_name) == 0) return(pcl);
		pcl = pcl->next;
	}
	return(NULL);
}

/******************************************************************/
/*
    FUNCTION NAME:

      tape_in_prodsys

  DESCRIPTION:

      Determines whether a tape is in a production system's list of tapes.

  FORMAL PARAMETERS:

      struct prod_sys *ps - Production system to search.

      char *dev_name - Tape drive to look for.

      char *cpu_name - CPU the tape can be accessed from.

  FUNCTION VALUE:

      struct prod_tape_list * - Returns NULL if the tape is not in the list.
		Returns a pointer to the record if the tape is in the list.

								  */
/******************************************************************/

struct prod_tape_list *tape_in_prodsys(ps,dev_name,cpu_name)
struct prod_sys *ps;
char *dev_name;
char *cpu_name;
{
	struct prod_tape_list *ptl;

	ptl = ps->tapes;
	while (ptl != NULL) {
		if (strcmp(ptl->tape->dev_name,dev_name) == 0)
			if (cpu_name_compare(ptl->tape->cpu->cpu_name,cpu_name) == 0) return(ptl);
		ptl = ptl->next;
	}
	return(NULL);
}

/******************************************************************/
/*
    FUNCTION NAME:

      change_jm_cpu

  DESCRIPTION:

      Changes the CPU on which the JM will run for a production system.

  FORMAL PARAMETERS:

      struct prod_sys *p - Production system to modify.

      struct cpu_list *c - CPU that will run the JM.

								  */
/******************************************************************/

change_jm_cpu(p,c)
struct prod_sys *p;
struct cpu_list *c;
{

	/* Update the current JM CPU */

	p->jm_cpu->refcount--;
	p->jm_cpu->np--;

	/* Change the pointer */

	p->jm_cpu = c;

	/* Update the new JM CPU */

	p->jm_cpu->refcount++;
	p->jm_cpu->np++;

}

/******************************************************************/
/*
    FUNCTION NAME:

      update_prodsys_status

  DESCRIPTION:

      Resets the status of each working or reduced production system
      in the list.  Broken production systems must have their status
      changed directly by the system manager.

  FORMAL PARAMETERS:

      struct prod_sys *p - First production system in the list.

								  */
/******************************************************************/

update_prodsys_status(p)
struct prod_sys *p;
{

	struct prod_cpu_list *pcl;
	struct prod_tape_list *ptl;

	int all_broken,all_working;

	while (p != NULL) {

		if (p->status == DB_BROKEN) {
			p = p->next;
			continue;
		}

		p->status = DB_WORKING;	/* working, full power */

		if (p->jm_cpu->status == DB_BROKEN) {
			p->status = DB_BROKEN;
			p = p->next;
			continue;
		}

		pcl = p->cpus;
		all_broken = -1;
		all_working = -1;
		while (pcl != NULL) {
			if (pcl->cpu->status == DB_WORKING)
				all_broken = 0;
			else if (pcl->cpu->status == DB_BROKEN)
				all_working = 0;
			pcl = pcl->next;
		}
		if (all_broken) {
			p->status = DB_BROKEN;
			p = p->next;
			continue;
		}
		else if (!all_working) {
			p->status = DB_REDUCED;
		}

		ptl = p->tapes;
		if (ptl == NULL) {	/* No tape drives */
			p = p->next;
			continue;
		}
		all_broken = -1;
		all_working = -1;
		while (ptl != NULL) {
			if (ptl->tape->status == DB_WORKING)
				all_broken = 0;
			else if (ptl->tape->status == DB_BROKEN)
				all_working = 0;
			ptl = ptl->next;
		}
		if (all_broken)
			p->status = DB_BROKEN;
		else if (!all_working)
			p->status = DB_REDUCED;

		p = p->next;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_word_list

  DESCRIPTION:

      Deallocates a word list structure.

  FORMAL PARAMETERS:

      struct word_list *wlist - First record in the list.

								  */
/******************************************************************/

clear_word_list(wlist)
struct word_list *wlist;
{
	struct word_list *w;

	while (wlist != NULL) {
		w = wlist->next;
		clear_word_list(wlist->down);
		if (wlist->word != NULL) {
			hfree(wlist->word);
		}
		hfree(wlist);
		wlist = w;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_ordered_tape_list

  DESCRIPTION:

      Frees the memory allocated for an ordered tape list.

  FORMAL PARAMETERS:

      struct ordered_tape_list *otl - Pointer to the first record in the list.

								  */
/******************************************************************/

clear_ordered_tape_list(otl)
struct ordered_tape_list *otl;
{
	struct ordered_tape_list *t;

	while (otl != NULL) {
		t = otl->next;
		hfree(otl->vsn);
		hfree(otl);
		otl = t;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_tape_list_heads

  DESCRIPTION:

      Deallocates a list of tape list headers.

  FORMAL PARAMETERS:

      struct tape_list_head *tlh - Pointer to the first header in the
				   list.

								  */
/******************************************************************/

clear_tape_list_heads(tlh)
struct tape_list_head *tlh;
{
	struct tape_list_head *h;

	while (tlh != NULL) {
		h = tlh->next;
		clear_ordered_tape_list(tlh->tlist);
		hfree(tlh);
		tlh = h;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_class_list

  DESCRIPTION:

      Deallocates a class list from a job.

  FORMAL PARAMETERS:

      struct class_list *clist - Pointer to the first record in the list.

								  */
/******************************************************************/

clear_class_list(clist)
struct class_list *clist;
{
	struct class_list *c;

	while (clist != NULL) {
		c = clist->next;
		clear_tape_list_heads(clist->tlh);
		clear_word_list(clist->scpu);
		clear_word_list(clist->exe);
		if (clist->stin != NULL) {
			hfree(clist->stin);
		}
		if (clist->stout != NULL) {
			hfree(clist->stout);
		}
		clear_word_list(clist->dfils);
		if (clist->outdir != NULL) {
			hfree(clist->outdir);
		}
		clear_word_list(clist->outfil);
		hfree(clist);
		clist = c;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      clear_prod_job_list

  DESCRIPTION:

      Deallocates a production system's list of jobs.

  FORMAL PARAMETERS:

      struct prod_job_list *pjl - Pointer to first record in the list.

								  */
/******************************************************************/

clear_prod_job_list(pjl)
struct prod_job_list *pjl;
{
	struct prod_job_list *j;

	while (pjl != NULL) {
		j = pjl->next;
		clear_class_list(pjl->clist);
		if (pjl->sub != NULL) {
			hfree(pjl->sub->jcf);
			hfree(pjl->sub->jdf);
			hfree(pjl->sub->username);
			hfree(pjl->sub->node);
			hfree(pjl->sub->time);
			hfree(pjl->sub);
		}
		if (pjl->setup != NULL) {
			hfree(pjl->setup);
		}
		if (pjl->logdir != NULL) {
			hfree(pjl->logdir);
		}
		if (pjl->on_deck_time != NULL) {
			hfree(pjl->on_deck_time);
		}
		if (pjl->start_time != NULL) {
			hfree(pjl->start_time);
		}
		if (pjl->rundown_time != NULL) {
			hfree(pjl->rundown_time);
		}
		clear_word_list(pjl->notif);
		clear_word_list(pjl->notif_s);
		clear_word_list(pjl->notif_f);
		hfree(pjl);
		pjl = j;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      create_job_record

  DESCRIPTION:

        Allocates a structure for a new CPS job and returns a pointer to that
      job.  Initializes all the fields in the job structure.

  FORMAL PARAMETERS:

       int entry
       int pipestage
       int status
       int area

  FUNCTION VALUE:

      (struct prod_job_list *) - Returns a pointer to the newly created record.

								  */
/******************************************************************/

struct prod_job_list *create_job_record(entry,pipestage,status,area)
int entry,pipestage,status,area;
{

	struct prod_job_list *j;

	int i,*x;

	j = (struct prod_job_list *) hmalloc(sizeof(struct prod_job_list));

	/* Initialize job information */

	j->entry = entry;
	j->pipestage = pipestage;
	j->status = status;
	j->area = area;
	j->abort = 0;
	j->jcss_pid = 0;

	j->sub = NULL;

	j->clist = NULL;

	j->on_deck_time = NULL;
	j->start_time = NULL;
	j->rundown_time = NULL;

	j->setup = NULL;
	j->cleanup = NULL;
	j->logdir = NULL;
	j->notif = NULL;
	j->notif_s = NULL;
	j->notif_f = NULL;
	j->next = NULL;

	return(j);
}

/******************************************************************/
/*
    FUNCTION NAME:

      assign_queue_priority

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

assign_queue_priority(p,j)
struct prod_sys *p;
struct prod_job_list *j;
{

	int high;
	struct prod_job_list *j1;

	high = 0;
	j1 = p->job_queue;
	while (j1 != NULL) {
		high = (high > j1->qpri) ? high : j1->qpri;	/* max(high,j1->qpri) */
		j1 = j1->next;
	}
	high++;

	j->qpri = high;
	return;
}

/******************************************************************/
/*
    FUNCTION NAME:

      sort_job_queue

  DESCRIPTION:

      Orders jobs in list from highest priority to lowest priority.

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

sort_job_queue(p)
struct prod_sys *p;
{
	struct prod_job_list *jq,*j0,*j1,*j2,*sj;

	sj = NULL;
	jq = p->job_queue;
	while (jq != NULL) {
		j2 = jq->next;
		j0 = NULL;
		j1 = sj;
		while ((j1 != NULL) && (j1->qpri < jq->qpri)) {  /* Find insertion pt */
			j0 = j1;
			j1 = j1->next;
		}
		if (j0 != NULL)
			j0->next = jq;
		else
			sj = jq;
		jq->next = j1;
		jq = j2;
	}

	/* Set pointer to last job in queue */

	if (sj == NULL)
		p->back = NULL;
	else {
		j0 = sj;
		while (j0->next != NULL) j0 = j0->next;
		p->back = j0;
	}

	p->job_queue = sj;
	return;
}

/******************************************************************/
/*
    FUNCTION NAME:

      extract_from_queue

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

struct prod_job_list *extract_from_queue(p,entnum)
struct prod_sys *p;
int entnum;
{
	struct prod_job_list *j0,*j1;

	j0 = NULL;
	j1 = p->job_queue;
	while ((j1 != NULL) && (j1->entry != entnum)) {
		j0 = j1;
		j1 = j1->next;
	}
	if (j1 == NULL) return(j1);	/* Could not find job */

	if (j0 == NULL)
		p->job_queue = j1->next;
	else
		j0->next = j1->next;

	if (j1 == p->back) p->back = j0;

	return(j1);
}

/******************************************************************/
/*
    FUNCTION NAME:

      get_highest_priority

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

int get_highest_priority(p)
struct prod_sys *p;
{
	struct prod_job_list *j;
	int high;

	j = p->job_queue;
	high = 999999;
	while (j != NULL) {
		high = (high < j->qpri) ? high : j->qpri;  /* max(high,j->qpri) */
		j = j->next;
	}
	return(high);
}

/******************************************************************/
/*
    FUNCTION NAME:

      get_priority

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

int get_priority(p,entnum)
struct prod_sys *p;
int entnum;
{
	struct prod_job_list *j;
	int high;

	j = p->job_queue;
	while (j != NULL) {
		if (entnum == j->entry) return(j->qpri);
		j = j->next;
	}
	return(-1);
}

/******************************************************************/
/*
    FUNCTION NAME:

      fill_submit_info

  DESCRIPTION:

      Fills in the submit sub-structure for the job in the newjob slot.

  FORMAL PARAMETERS:

       struct prod_sys *p
       char *jcf
       char *jdf
       char *username
       char *node
       char *time

								  */
/******************************************************************/

fill_submit_info(p,jcf,jdf,username,node,tim,addr)
struct prod_sys *p;
char *jcf,*jdf,*username,*node,*tim,*addr;
{

	struct prod_job_list *j;
	struct submit_info *s;

	j = p->newjob;
	if (j == NULL) return;

	s = (struct submit_info *) hmalloc(sizeof(struct submit_info));
	j->sub = s;

	/* Initialize the submit information */

	s->jcf = (char *) hmalloc(strlen(jcf)+1);
	strcpy(s->jcf,jcf);

	s->jdf = (char *) hmalloc(strlen(jdf)+1);
	strcpy(s->jdf,jdf);

	s->node = (char *) hmalloc(strlen(node)+1);
	strcpy(s->node,node);

	s->time = (char *) hmalloc(strlen(tim)+1);
	strcpy(s->time,tim);

	s->username = (char *) hmalloc(strlen(username)+1);
	strcpy(s->username,username);

	s->address = (char *) hmalloc(strlen(addr)+1);
	strcpy(s->address,addr);
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_class

  DESCRIPTION:

      Adds a class to a job's class list if it is not already there.  Returns
      a pointer to the class record.

  FORMAL PARAMETERS:

      struct prod_job_list *j - Pointer to the job structure.

      int c - Class number to be added.

  FUNCTION VALUE:

      (struct class_list *) - Returns a pointer to the class record.

								  */
/******************************************************************/

struct class_list *add_class(j,c)
struct prod_job_list *j;
int c;
{

	struct class_list *cl;

	/* Search for the class */

	cl = j->clist;
	while (cl != NULL)
		if (cl->class == c)
			return(cl);
		else
			cl = cl->next;

	/* Class was not found, create a new one and initialize the record */

	cl = (struct class_list *) hmalloc(sizeof(struct class_list));

	cl->class = c;
	cl->opsys = DB_UNKNOWN;
	cl->tlh = NULL;		/* Ordered tape lists */
	cl->debug = 0;
	cl->minproc = 1;
	cl->numproc = 1;
	cl->percpu = -1;
	cl->ptimeout = -1;
	cl->ctype = -1;
	cl->scpu = NULL;
	cl->share = -1;
	cl->taptyp[0] = -1;
	cl->exe = NULL;
	cl->stin = NULL;
	cl->stout = NULL;
	cl->dfils = NULL;
	cl->outdir = NULL;
	cl->outfil = NULL;

	/* Add the class to the list */

	cl->next = j->clist;
	j->clist = cl;

	return(cl);

}

/******************************************************************/
/*
    FUNCTION NAME:

      find_class

  DESCRIPTION:

      Finds a class in a job's class list.

  FORMAL PARAMETERS:

      struct prod_job_list *j - Pointer to the job structure.

      int c - Class number to be added.

  FUNCTION VALUE:

      (struct class_list *) - Returns a pointer to the class record.
			      (NULL if not found).

								  */
/******************************************************************/

struct class_list *find_class(j,c)
struct prod_job_list *j;
int c;
{

	struct class_list *cl;

	/* Search for the class */

	cl = j->clist;
	while (cl != NULL)
		if (cl->class == c)
			return(cl);
		else
			cl = cl->next;

	return(cl);

}

/******************************************************************/
/*
    FUNCTION NAME:

      add_tape_list

  DESCRIPTION:

      Adds a tape list head to a class's list of tape lists if it doesn't
      already exist.  Fields that are specified (not DB_NOPARAM) are filled
      in.

  FORMAL PARAMETERS:

      struct class_list *cl - Pointer to the class.

      int ln - The tape list number.  If DB_NOPARAM is specified, then the
	       list number defaults to the first tape list in the list.
	       If there are no tape lists, the list number defaults to 1.

      int io - Input or Output (DB_INPUT or DB_OUTPUT).  If DB_NOPARAM 
	       is specified, the io field is not changed.

      int dt - Device type.  If DB_NOPARAM is specified, the devtype field
	       is not changed.

  FUNCTION VALUE:

      (struct tape_list_head *) - Pointer to the tape list that was created
				  modified, or found.

								  */
/******************************************************************/

struct tape_list_head *add_tape_list(cl,ln,io,dt)
struct class_list *cl;
int ln,io,dt;
{

	struct tape_list_head *tlh;

	/* Specify the list number of DB_NOPARAM was used */

	if (ln == DB_NOPARAM) {
		if (cl->tlh == NULL) {
			ln = 1;
			goto addlist;
		}
		else {
			ln = cl->tlh->listnum;
			tlh = cl->tlh;
			goto modlist;
		}

	}

	/* Find the tape list */

	tlh = cl->tlh;
	while (tlh != NULL) {
		if (tlh->listnum == ln) goto modlist;
		tlh = tlh->next;
	}

	/* Add a new tape list */

addlist:
	tlh = (struct tape_list_head *) hmalloc(sizeof(struct tape_list_head));
	tlh->listnum = ln;
	tlh->devtype = DB_NOPARAM;
	tlh->io = DB_NOPARAM;
	tlh->num_tapes = 0;
	tlh->tlist = NULL;
	tlh->next = cl->tlh;
	cl->tlh = tlh;

	/* Modify the fields of the tape list head */

modlist:
	if (io != DB_NOPARAM) tlh->io = io;
	if (dt != DB_NOPARAM) tlh->devtype = dt;

	return(tlh);

}

/******************************************************************/
/*
    FUNCTION NAME:

      add_word_to_list

  DESCRIPTION:

      Adds a word to a word list and returns the new pointer to the start
      of the list.

  FORMAL PARAMETERS:

      struct word_list *wl - Word list.

      char *w - Word to add.

  FUNCTION VALUE:

      (struct word_list *) - New beginning of the list.

								  */
/******************************************************************/

struct word_list *add_word_to_list(wl,w)
struct word_list *wl;
char *w;
{

	struct word_list *nw,*wl1;

	nw = (struct word_list *) hmalloc(sizeof(struct word_list));
	nw->word = (char *) hmalloc(strlen(w)+1);
	strcpy(nw->word,w);
	nw->down = NULL;
	nw->next = NULL;

	if (wl == NULL)
		wl = nw;
	else {
		wl1 = wl;
		while (wl1->next != NULL) wl1 = wl1->next;
		wl1->next = nw;
	}

	return(wl);

}

/******************************************************************/
/*
    FUNCTION NAME:

      find_word

  DESCRIPTION:

      Locates a word in a word list and returns a pointer to it.

  FORMAL PARAMETERS:

      struct word_list *wl - Word list.

      char *w - Word to add.

  FUNCTION VALUE:

      (struct word_list *) - Pointer to word if found, NULL otherwise.

								  */
/******************************************************************/

struct word_list *find_word(wl,w)
struct word_list *wl;
char *w;
{
	while (wl != NULL) {
		if (strcmp(w,wl->word) == 0) return(wl);
		wl = wl->next;
	}
	return(wl);
}

/******************************************************************/
/*
    FUNCTION NAME:

      remove_tape_from_list

  DESCRIPTION:

      Removes all occurances of a tape from a job's tape list.

  FORMAL PARAMETERS:

      struct tape_list_head *tlh - List to remove tape from.

      char *vsn - Volume serial number of the tape.

								  */
/******************************************************************/

remove_tape_from_list(tlh,vsn)
struct tape_list_head *tlh;
char *vsn;
{

	struct ordered_tape_list *otl,*otl1;

	otl = tlh->tlist;
	otl1 = NULL;
	while (otl != NULL) {
		if (strcmp(otl->vsn,vsn) == 0) {
			if (otl1 == NULL) {
				tlh->tlist = otl->next;
				hfree(otl->vsn);
				hfree(otl);
				otl1 = NULL;
				otl = tlh->tlist;
			}
			else {
				otl1->next = otl->next;
				hfree(otl->vsn);
				hfree(otl);
				otl = otl1->next;
			}
			tlh->num_tapes--;
		}
		else {
			otl1 = otl;
			otl = otl->next;
		}
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_tape_to_list

  DESCRIPTION:

      Adds a tape to a tape list.

  FORMAL PARAMETERS:

      struct tape_list_head *tlh - Tape list to add tapes to.

      char *vsn - VSN of the tape to add.

  FUNCTION VALUE:

      none

								  */
/******************************************************************/

add_tape_to_list(tlh,vsn)
struct tape_list_head *tlh;
char *vsn;
{

	struct ordered_tape_list *t,*t1;

	/* Allocate a tape record and fill in the contents */

	t = (struct ordered_tape_list *) hmalloc(sizeof(struct ordered_tape_list));
	t->vsn = (char *) hmalloc(strlen(vsn)+1);
	strcpy(t->vsn,vsn);
	t->time_req = 0;
	t->time_mnt = 0;
	t->time_done = 0;
	t->status = DB_UNPROCESSED;
	t->next = NULL;

	/* Append the tape to the end of the list */

	if (tlh->tlist == NULL) {
		tlh->tlist = t;
	}
	else {
		t1 = tlh->tlist;
		while(t1->next != NULL) t1 = t1->next;
		t1->next = t;
	}
	tlh->num_tapes++;
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_tapes_to_list

  DESCRIPTION:

      Adds a list of tapes to a tape list.

  FORMAL PARAMETERS:

      struct tape_list_head *tlh - Tape list to add tapes to.

      struct word_list *wl - List of tapes to add.

  FUNCTION VALUE:

      none

  SIDE EFFECTS:

      The memory used for the input word list is deallocated.
								  */
/******************************************************************/

add_tapes_to_list(tlh,wl)
struct tape_list_head *tlh;
struct word_list *wl;
{

	struct word_list *wl1;

	while (wl != NULL) {
		wl1 = wl->next;
		add_tape_to_list(tlh,wl->word);
		hfree(wl->word);
		hfree(wl);
		wl = wl1;
	}

}

/******************************************************************/
/*
    FUNCTION NAME:

      add_tape_type

  DESCRIPTION:

      Adds a tape requirement to a class record.

  FORMAL PARAMETERS:

      struct class_list *cl - Class to modify.

      int taptyp - Type of tape to add.

      int numreq - Number of tape drives required.

								  */
/******************************************************************/

add_tape_type(cl,taptyp,numreq)
struct class_list *cl;
int taptyp,numreq;
{

	int i;

	i = 0;
	while (cl->taptyp[i] != -1) i++;

	cl->taptyp[i] = taptyp;
	cl->tapnum[i] = numreq;
	cl->taptyp[++i] = -1;
}

/******************************************************************/
/*
    FUNCTION NAME:

      parse_prog_line

  DESCRIPTION:

      Takes a list of words used as a program specification and fills the
      exe, stin, and stout fields of a class structure.

  FORMAL PARAMETERS:

      struct class_list *cl - Class to modify.

      struct word_list *w - List of words.

								  */
/******************************************************************/

parse_prog_line(cl,w)
struct class_list *cl;
struct word_list *w;
{

	struct word_list *cw,*pw,*w1;

	/* Remove input & output redirection */

	cw = w;
	pw = NULL;
	while (cw != NULL) {
		if (cw->word[0] == '>') {
			cl->stout = cw->word;
			if (pw == NULL) {
				w = cw->next;
				hfree(cw);
				cw = w;
			}
			else {
				pw->next = cw->next;
				hfree(cw);
				cw = pw->next;
			}
		}
		else if (cw->word[0] == '<') {
			cl->stin = cw->word;
			if (pw == NULL) {
				w = cw->next;
				hfree(cw);
				cw = w;
			}
			else {
				pw->next = cw->next;
				hfree(cw);
				cw = pw->next;
			}
		}
		else {
			pw = cw;
			cw = cw->next;
		}
	}

	/* Separate programs and arguments */

	while (w != NULL) {

		/* Find next pipeline or end of list */

		cw = w;
		pw = NULL;
		while ((cw != NULL) && (cw->word[0] != '|')) {
			pw = cw;
			cw = cw->next;
		}

		if (pw != NULL) pw->next = NULL;

		w->down = w->next;
		w->next = NULL;

		w1 = cl->exe;
		if (w1 == NULL)
			cl->exe = w;
		else {
			while (w1->next != NULL) w1 = w1->next;
			w1->next = w;
		}

		if (cw != NULL) {
			w = cw->next;
			hfree(cw->word);
			hfree(cw);
		}
		else
			w = NULL;
	}
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_job

  DESCRIPTION:

      Locates a job on a specified production system.  Returns a pointer to
      the job or NULL if not found.

  FORMAL PARAMETERS:

      struct prod_sys *p - Pointer to the production system.

      int pipestage - Stage of the pipe the job is located in.

      int entry - Entry number of the job.

  FUNCTION VALUE:

      (struct prod_job_list *) - Pointer to the job if found, otherwise NULL.

								  */
/******************************************************************/

struct prod_job_list *find_job(p,pipestage,entry)
struct prod_sys *p;
int pipestage,entry;
{

	struct prod_job_list *j;

	switch (pipestage) {

	case DB_PIPE_NEW :
		if (p->newjob == NULL) return(NULL);
		if (p->newjob->entry == entry) return(p->newjob);
		break;
	case DB_PIPE_QUEUE :
		j = p->job_queue;
		while (j != NULL) {
			if (j->entry == entry) return(j);
			j = j->next;
		}
		break;
	case DB_PIPE_ON_DECK :
		if (p->on_deck == NULL) return(NULL);
		if (p->on_deck->entry == entry) return(p->on_deck);
		break;
	case DB_PIPE_CURRENT :
		if (p->current == NULL) return(NULL);
		if (p->current->entry == entry) return(p->current);
		break;
	case DB_PIPE_RUNDOWN :
		if (p->rundown == NULL) return(NULL);
		if (p->rundown->entry == entry) return(p->rundown);
		break;
	}
	return(NULL);
}

/******************************************************************/
/*
    FUNCTION NAME:

      find_tape_list

  DESCRIPTION:

      Locates an ordered tape list within a specific job.

  FORMAL PARAMETERS:

      struct prod_job_list *j - Job to search.

      int class - Class number.

      int listnum - Tape list number.

  FUNCTION VALUE:

      (struct tape_list_head *) - Pointer to the ordered tape list, NULL
				  if not found.

								  */
/******************************************************************/

struct tape_list_head *find_tape_list(j,class,listnum)
struct prod_job_list *j;
int class,listnum;
{

	struct class_list *c;
	struct tape_list_head *tlh;

	c = j->clist;
	while (c != NULL) {
		if (c->class == class) {
			tlh = c->tlh;
			while (tlh != NULL) {
				if (tlh->listnum == listnum) return(tlh);
				tlh = tlh->next;
			}
			return(NULL);
		}
		c = c->next;
	}
	return(NULL);
}

/******************************************************************/
/*
    FUNCTION NAME:

      move_to_front

  DESCRIPTION:

      Moves a job from anywhere in the queue to the front of the queue.

  FORMAL PARAMETERS:

      struct prod_sys *p - Production system.

      struct prod_job_list *j - Job to move to the front.

								  */
/******************************************************************/

move_to_front(p,j)
struct prod_sys *p;
struct prod_job_list *j;
{

	struct prod_job_list *j1,*j2;

	j1 = NULL;
	j2 = p->job_queue;
	while (j2 != j) {
		if (j2 == NULL) return;	    /* Job not found, do nothing */
		j1 = j2;
		j2 = j2->next;
	}

	if (j1 == NULL) {
		return;	    /* Already at front */
	}

	j1->next = j2->next;
	if (p->back == j2) p->back = j1;
	j2->next = p->job_queue;
	p->job_queue = j2;
}

/******************************************************************/
/*
    FUNCTION NAME:

      delete_queue_entry

  DESCRIPTION:

      Removes an entry from a queue.

  FORMAL PARAMETERS:

      struct prod_sys *p - Production system.

      int e - Entry number of the job to be removed.

								  */
/******************************************************************/

delete_queue_entry(p,e)
struct prod_sys *p;
int e;
{

	struct prod_job_list *j1,*j2;

	/* Find the job on the queue */

	j1 = NULL;
	j2 = p->job_queue;
	while ((j2 != NULL) && (j2->entry != e)) {
		j1 = j2;
		j2 = j2->next;
	}
	if (j2 == NULL) return;

	/* Remove the job from the queue */

	if (j1 == NULL)
		p->job_queue = j2->next;
	else
		j1->next = j2->next;

	if (p->back == j2) p->back = j1;

	/* Deallocate the memory */

	j2->next = NULL;
	clear_prod_job_list(j2);

}

/******************************************************************/
/*
    FUNCTION NAME:

      find_tape_reel

  DESCRIPTION:

      Locates a tape within a tape list (or returns NULL).

  FORMAL PARAMETERS:

      struct tape_list_head *tlh - Tape list to search.

      char *vsn - VSN of the tape to look for.

  FUNCTION VALUE:

      (struct ordered_tape_list *) - Pointer to the tape reel record.

								  */
/******************************************************************/

struct ordered_tape_list *find_tape_reel(tlh,vsn)
struct tape_list_head *tlh;
char *vsn;
{
	struct ordered_tape_list *otl;

	otl = tlh->tlist;
	while (otl != NULL) {
		if (strcmp(otl->vsn,vsn) == 0) return(otl);
		otl = otl->next;
	}
	return(otl);
}

/******************************************************************/
/*
    FUNCTION NAME:

      insert_tape_in_list

  DESCRIPTION:

      Inserts a tape reel into an ordered tape list.

  FORMAL PARAMETERS:

      struct tape_list_head *tlh - Tape list header.

      char *vsn - VSN of the tape reel to insert after.  "*" means insert
		  at the beginning of the list.

      char *new_vsn - VSN of the new tape to insert.

								  */
/******************************************************************/

insert_tape_in_list(tlh,vsn,new_vsn)
struct tape_list_head *tlh;
char *vsn,*new_vsn;
{
	struct ordered_tape_list *t,*t1;

	/* Allocate a tape record and fill in the contents */

	t = (struct ordered_tape_list *) hmalloc(sizeof(struct ordered_tape_list));
	t->vsn = (char *) hmalloc(strlen(new_vsn)+1);
	strcpy(t->vsn,new_vsn);
	t->status = DB_UNPROCESSED;

	if (vsn[0] == '*') {    /* Insert tape at front of the list */
		t->next = tlh->tlist;
		tlh->tlist = t;
		return;
	}

	t1 = find_tape_reel(tlh,vsn);
	if (t1 == NULL) {
		hfree(t);
		return;
	}

	t->next = t1->next;
	t1->next = t;
	tlh->num_tapes++;
	return;
}

/******************************************************************/
/*
    FUNCTION NAME:

      database_update

  DESCRIPTION:

      All messages received from ALL types of clients which modify the
      database are sent here.  All modifications to the PM's database are
      done from this routine.

  FORMAL PARAMETERS:

      struct fs_list **fs - The first entry point into the database -
			        the pointer to the list of disks.

      struct cpu_list **cpu - The second entry point into the database -
			      the pointer to the list of CPUs.

      struct tape_drive_list **tape - The third entry point into the database -
				      the pointer to the list of tapes.

      struct prod_sys **ps - The last entry point into the database -
			     the pointer to the list of production systems.

      char **args - The arguments of a "db_update" message.  The message
		    must have already been parsed to determine it was a
		    "db_update" message to begin with.

    FUNCTION VALUE:

      (int) - 0 if more update messages are coming, -1 if the DB_NO_MORE_UPDATES
	      message came or an unrecognizable messsage arrived, 1 if the
	      database was not updated due to an inconsistency.


								  */
/******************************************************************/

int database_update(fs,cpu,tape,ps,args)
struct fs_list **fs;
struct cpu_list **cpu;
struct tape_drive_list **tape;
struct prod_sys **ps;
char **args;
{
	struct fs_list *f;
	struct cpu_list *c;
	struct tape_drive_list *t;
	struct prod_sys *p;
	struct prod_cpu_list *pcl;
	struct prod_job_list *j;
	struct class_list *cl;
	struct tape_list_head *tlh;
	struct ordered_tape_list *otl;
	struct word_list *wl1,*wl2;

	int msgtyp;

	int mstate;

	int pipestage;

	int entnum;

	int incon_no;
	char incon_msg[80];

	int i;

	/* Next inconsistency number = 97 */

	msgtyp = atoi(args[0]);

	switch (msgtyp) {
	case DB_ADD_CPU :
		if (args[1] == NULL) {
			incon_no = 91;
			goto inconsistent;
		}
		c = add_cpu(cpu,args[1],atoi(args[2]),atoi(args[4]));
		c->status = atoi(args[3]);
		c->task = atoi(args[5]);
		break;
	case DB_MODIFY_CPU :
		c = find_cpu(*cpu,args[1]);
		if (c == NULL) {
			incon_no = 1;
			goto inconsistent;
		}
		c->cpu_type = atoi(args[2]);
		c->opsys = DB_UNIX;
		c->status = atoi(args[3]);
		c->power = atoi(args[4]);
		c->task = atoi(args[5]);
		break;
	case DB_DELETE_CPU :
		delete_cpu(cpu,args[1]);
		break;
	case DB_ADD_TAPE :
		c = find_cpu(*cpu,args[3]);
		if (c == NULL) {
			incon_no = 2;
			goto inconsistent;
		}
		t = add_tape(tape,args[1],args[6],atoi(args[2]),c);
		t->status = atoi(args[4]);
		t->used = atoi(args[5]);
		break;
	case DB_MODIFY_TAPE :
		t = find_tape(*tape,args[1],args[3]);
		if (t == NULL) {
			incon_no = 3;
			goto inconsistent;
		}
		c = find_cpu(*cpu,args[3]);
		if (c == NULL) {
			incon_no = 4;
			goto inconsistent;
		}
		t->dev_type = atoi(args[2]);
		t->cpu = c;
		t->status = atoi(args[4]);
		t->used = atoi(args[5]);
		break;
	case DB_DELETE_TAPE :
		delete_tape(tape,args[1],args[2]);
		break;
	case DB_ADD_PRODSYS :
		c = find_cpu(*cpu,args[2]);
		if (c == NULL) {
			incon_no = 5;
			goto inconsistent;
		}
		f = find_fs(*fs,atoi(args[8]));
		if (f == NULL) {
			incon_no = 6;
			goto inconsistent;
		}
		p = add_prodsys(ps,args[1],c);
		p->status = atoi(args[3]);
		p->next_entry = atoi(args[4]);
		p->account = (char *) hmalloc(strlen(args[5])+1);
		strcpy(p->account,args[5]);
		p->passwd = (char *) hmalloc(strlen(args[6])+1);
		strcpy(p->passwd,args[6]);
		p->login_dir = (char *) hmalloc(strlen(args[7])+1);
		strcpy(p->login_dir,args[7]);
		p->jm_fs = f;
		f->refcount++;
		break;
	case DB_MODIFY_PRODSYS :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 7;
			goto inconsistent;
		}
		c = find_cpu(*cpu,args[2]);
		if (c == NULL) {
			incon_no = 8;
			goto inconsistent;
		}
		f = find_fs(*fs,atoi(args[8]));
		if (f == NULL) {
			incon_no = 9;
			goto inconsistent;
		}
		change_jm_cpu(p,c);
		p->jm_fs->refcount--;
		p->jm_fs = f;
		f->refcount++;
		p->status = atoi(args[3]);
		p->next_entry = atoi(args[4]);
		if (args[5][0] != '-') {
			hfree(p->account);
			p->account = (char *) hmalloc(strlen(args[5])+1);
			strcpy(p->account,args[5]);
		}
		if (args[6][0] != '-') {
			hfree(p->passwd);
			p->passwd = (char *) hmalloc(strlen(args[6])+1);
			strcpy(p->passwd,args[6]);
		}
		if (args[7][0] != '-') {
			hfree(p->login_dir);
			p->login_dir = (char *) hmalloc(strlen(args[7])+1);
			strcpy(p->login_dir,args[7]);
		}
		break;
	case DB_DELETE_PRODSYS :
		delete_prodsys(ps,args[1]);
		break;
	case DB_ADD_CPU_TO_PROD :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 10;
			goto inconsistent;
		}
		c = find_cpu(*cpu,args[2]);
		if (c == NULL) {
			incon_no = 11;
			goto inconsistent;
		}
		f = find_fs(*fs,atoi(args[4]));
		if (f == NULL) {
			incon_no = 12;
			goto inconsistent;
		}
		add_cpu_to_prod(p,c,f,atoi(args[3]),args[5],atoi(args[6]));
		break;
	case DB_MODIFY_CPU_IN_PROD :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 13;
			goto inconsistent;
		}
		pcl = cpu_in_prodsys(p,args[2]);
		if (pcl == NULL) {
			incon_no = 14;
			goto inconsistent;
		}
		f = find_fs(*fs,atoi(args[4]));
		if (f == NULL) {
			incon_no = 15;
			goto inconsistent;
		}
		if (pcl->fs != NULL) pcl->fs->refcount--;
		pcl->fs = f;
		f->refcount++;
		pcl->max_proc = atoi(args[3]);
		pcl->lognum = atoi(args[6]);
		break;
	case DB_DELETE_CPU_FROM_PROD :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 16;
			goto inconsistent;
		}
		c = find_cpu(*cpu,args[2]);
		if (c == NULL) {
			incon_no = 17;
			goto inconsistent;
		}
		delete_cpu_from_prod(p,c);
		break;
	case DB_ADD_TAPE_TO_PROD :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 18;
			goto inconsistent;
		}
		t = find_tape(*tape,args[2],args[3]);
		if (t == NULL) {
			incon_no = 19;
			goto inconsistent;
		}
		add_tape_to_prod(p,t);
		break;
	case DB_DELETE_TAPE_FROM_PROD :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 20;
			goto inconsistent;
		}
		t = find_tape(*tape,args[2],args[3]);
		if (t == NULL) {
			incon_no = 21;
			goto inconsistent;
		}
		delete_tape_from_prod(p,t);
		break;
	case DB_ADD_FILESYSTEM :
		c = find_cpu(*cpu,args[1]);
		if (c == NULL) {
			incon_no = 22;
			goto inconsistent;
		}
		add_fs(fs,c,args[2],atoi(args[3]),atoi(args[4]));
		break;
	case DB_REMOVE_FILESYSTEM :
		remove_fs(fs,atoi(args[1]));
		break;
	case DB_ADD_JOB :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 25;
			goto inconsistent;
		}
		if (p->newjob != NULL) clear_prod_job_list(p->newjob);
		p->newjob = create_job_record(atoi(args[2]),atoi(args[3]),
		    atoi(args[4]),atoi(args[5]));
		assign_queue_priority(p,p->newjob);
		break;
	case DB_FILL_SUBMIT_INFO :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 26;
			goto inconsistent;
		}
		if (p->newjob == NULL) {
			incon_no = 27;
			goto inconsistent;
		}
		fill_submit_info(p,args[2],args[3],args[4],args[5],args[6],
		    args[7]);
		break;
	case DB_ADVANCE_JOB :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 28;
			goto inconsistent;
		}
		switch (atoi(args[2])) {
		case DB_PIPE_QUEUE :
			if (p->on_deck != NULL) {
				incon_no = 29;
				goto inconsistent;
			}
			if (p->job_queue == NULL) {
				incon_no = 30;
				goto inconsistent;
			}
			entnum = atoi(args[4]);
			if (get_highest_priority(p) != get_priority(p,entnum)) {
				printerr(FACILITY,ERROR_MSG,
				    "Attempted to advance low priority job from queue",
				    NULL,1);
				return(1);
			}
			j = extract_from_queue(p,entnum);
			if (j == NULL) {
				printerr(FACILITY,ERROR_MSG,
				    "Failed to extract job from queue",
				    NULL,2);
				return(1);
			}
			p->on_deck = j;
			j->next = NULL;
			j->pipestage = DB_PIPE_ON_DECK;
			j->status = DB_IN_PROGRESS;
			j->on_deck_time = 
			    (char *) hmalloc(strlen(args[3])+1);
			strcpy(j->on_deck_time,args[3]);
			break;
		case DB_PIPE_ON_DECK :
			if (p->current != NULL) {
				incon_no = 31;
				goto inconsistent;
			}
			if (p->on_deck == NULL) {
				incon_no = 32;
				goto inconsistent;
			}
			p->current = p->on_deck;
			p->on_deck = NULL;
			p->current->pipestage = DB_PIPE_CURRENT;
			p->current->status = DB_IN_PROGRESS;
			p->current->start_time = 
			    (char *) hmalloc(strlen(args[3])+1);
			strcpy(p->current->start_time,args[3]);
			break;
		case DB_PIPE_CURRENT :
			if (p->rundown != NULL) {
				incon_no = 33;
				goto inconsistent;
			}
			if (p->current == NULL) {
				incon_no = 34;
				goto inconsistent;
			}
			p->rundown = p->current;
			p->current = NULL;
			p->rundown->pipestage = DB_PIPE_RUNDOWN;
			p->rundown->status = DB_IN_PROGRESS;
			p->rundown->rundown_time = 
			    (char *) hmalloc(strlen(args[3])+1);
			strcpy(p->rundown->rundown_time,args[3]);
			break;
		case DB_PIPE_RUNDOWN :
			if (p->rundown != NULL) {
				clear_prod_job_list(p->rundown);
				p->rundown = NULL;
			}
			break;
		}
		break;
	case DB_REMOVE_JOB :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 35;
			goto inconsistent;
		}
		switch (atoi(args[2])) {
		case DB_PIPE_NEW :
			if (p->newjob != NULL) {
				clear_prod_job_list(p->newjob);
				p->newjob = NULL;
			}
			break;
		case DB_PIPE_QUEUE :
			delete_queue_entry(p,atoi(args[3]));
			break;
		case DB_PIPE_ON_DECK :
			if (p->on_deck != NULL) {
				p->job_area[p->on_deck->area-1] = 0;
				clear_prod_job_list(p->on_deck);
				p->on_deck = NULL;
			}
			break;
		case DB_PIPE_CURRENT :
			if (p->current != NULL) {
				p->job_area[p->current->area-1] = 0;
				clear_prod_job_list(p->current);
				p->current = NULL;
			}
			break;
		case DB_PIPE_RUNDOWN :
			if (p->rundown != NULL) {
				p->job_area[p->rundown->area-1] = 0;
				clear_prod_job_list(p->rundown);
				p->rundown = NULL;
			}
			break;
		}
		break;
	case DB_SET_JOB_AREA :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 36;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		j->area = atoi(args[4]);
		break;
	case DB_UPDATE_JOB_STATUS :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 38;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		j->status = atoi(args[4]);
		break;
	case DB_ADD_JOB_CLASS :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 40;
			goto inconsistent;
		}
		if (p->newjob == NULL) {
			incon_no = 41;
			goto inconsistent;
		}
		cl = add_class(p->newjob,atoi(args[2]));
		break;
	case DB_ADD_TAPE_LIST :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 42;
			goto inconsistent;
		}
		if (p->newjob == NULL) {
			incon_no = 43;
			goto inconsistent;
		}
		cl = find_class(p->newjob,atoi(args[2]));	/* Find the class */
		if (cl == NULL) {
			incon_no = 44;
			goto inconsistent;
		}
		tlh = add_tape_list(cl,atoi(args[3]),atoi(args[4]),atoi(args[5]));
		break;
	case DB_ADD_TAPE_REEL :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 45;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		tlh = find_tape_list(j,atoi(args[4]),atoi(args[5]));
		if (tlh == NULL) {
			incon_no = 47;
			goto inconsistent;
		}
		add_tape_to_list(tlh,args[6]);
		break;
	case DB_INSERT_TAPE_REEL :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 48;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		tlh = find_tape_list(j,atoi(args[4]),atoi(args[5]));
		if (tlh == NULL) {
			incon_no = 50;
			goto inconsistent;
		}
		insert_tape_in_list(tlh,args[6],args[7]);
		break;
	case DB_REMOVE_TAPE_REEL :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 51;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		tlh = find_tape_list(j,atoi(args[4]),atoi(args[5]));
		if (tlh == NULL) {
			incon_no = 53;
			goto inconsistent;
		}
		remove_tape_from_list(tlh,args[6]);
		break;
	case DB_TAPE_REEL_STATUS :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 54;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		tlh = find_tape_list(j,atoi(args[4]),atoi(args[5]));
		if (tlh == NULL) {
			incon_no = 56;
			goto inconsistent;
		}
		otl = find_tape_reel(tlh,args[6]);
		if (otl == NULL) {
			incon_no = 57;
			goto inconsistent;
		}
		otl->status = atoi(args[7]);
		break;
	case DB_MOVE_JOB_TO_QUEUE :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 58;
			goto inconsistent;
		}
		j = p->newjob;
		if (j == NULL) return(1);
		if (p->job_queue == NULL) {
			p->back = j;
			p->job_queue = j;
		}
		else {
			p->back->next = j;
			p->back = j;
		}
		p->newjob = NULL;
		j->pipestage = DB_PIPE_QUEUE;
		break;
	case DB_MOVE_JOB_TO_ON_DECK :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 60;
			goto inconsistent;
		}
		j = p->newjob;
		if (j == NULL) return(1);
		if (p->on_deck != NULL) clear_prod_job_list(p->on_deck);
		p->on_deck = j;
		p->newjob = NULL;
		j->pipestage = DB_PIPE_ON_DECK;
		break;
	case DB_MOVE_JOB_TO_CURRENT :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 62;
			goto inconsistent;
		}
		j = p->newjob;
		if (j == NULL) return(1);
		if (p->current != NULL) clear_prod_job_list(p->current);
		p->current = j;
		p->newjob = NULL;
		j->pipestage = DB_PIPE_CURRENT;
		break;
	case DB_MOVE_JOB_TO_RUNDOWN :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 64;
			goto inconsistent;
		}
		j = p->newjob;
		if (j == NULL) return(1);
		if (p->rundown != NULL) clear_prod_job_list(p->rundown);
		p->rundown = j;
		p->newjob = NULL;
		j->pipestage = DB_PIPE_RUNDOWN;
		break;
	case DB_ALLOCATE_AREA :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 66;
			goto inconsistent;
		}
		p->job_area[atoi(args[2])-1] = -1;
		break;
	case DB_FREE_AREA :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 67;
			goto inconsistent;
		}
		p->job_area[atoi(args[2])-1] = 0;
		break;
	case DB_JCSS_PID :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 94;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		j->jcss_pid = atoi(args[4]);
		break;
	case DB_SET_ABORT_FLAG :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 68;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		j->abort = -1;
		break;
	case DB_CLEAR_ABORT_FLAG :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 70;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		j->abort = 0;
		break;
	case DB_SET_PHASE_TIMES :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 72;
			goto inconsistent;
		}
		j = p->newjob;
		if (strcmp(args[2],"null") != 0) {
			j->on_deck_time = (char *) hmalloc(strlen(args[2])+1);
			strcpy(j->on_deck_time,args[2]);
		}
		if (strcmp(args[3],"null") != 0) {
			j->start_time = (char *) hmalloc(strlen(args[3])+1);
			strcpy(j->start_time,args[3]);
		}
		if (strcmp(args[4],"null") != 0) {
			j->rundown_time = (char *) hmalloc(strlen(args[4])+1);
			strcpy(j->rundown_time,args[4]);
		}
		break;
	case DB_MOUNT_INFO :
		t = find_tape(*tape,args[2],args[1]);
		if (t == NULL) {
			incon_no = 73;
			goto inconsistent;
		}

		strcpy(t->mi->vsn,args[3]);
		t->mi->rstate = atoi(args[4]);
		t->mi->mstate = atoi(args[5]);
		break;
	case DB_TAPE_TIME_REQ :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 77;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		tlh = find_tape_list(j,atoi(args[4]),atoi(args[5]));
		if (tlh == NULL) {
			incon_no = 79;
			goto inconsistent;
		}
		otl = find_tape_reel(tlh,args[6]);
		if (otl == NULL) {
			incon_no = 80;
			goto inconsistent;
		}
		otl->time_req = atol(args[7]);
		break;
	case DB_TAPE_TIME_MNT :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 81;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		tlh = find_tape_list(j,atoi(args[4]),atoi(args[5]));
		if (tlh == NULL) {
			incon_no = 83;
			goto inconsistent;
		}
		otl = find_tape_reel(tlh,args[6]);
		if (otl == NULL) {
			incon_no = 84;
			goto inconsistent;
		}
		otl->time_mnt = atol(args[7]);
		break;
	case DB_TAPE_TIME_DONE :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 85;
			goto inconsistent;
		}
		j = find_job(p,atoi(args[2]),atoi(args[3]));
		if (j == NULL) return(1);
		tlh = find_tape_list(j,atoi(args[4]),atoi(args[5]));
		if (tlh == NULL) {
			incon_no = 87;
			goto inconsistent;
		}
		otl = find_tape_reel(tlh,args[6]);
		if (otl == NULL) {
			incon_no = 88;
			goto inconsistent;
		}
		otl->time_done = atol(args[7]);
		break;
	case DB_ADD_USER :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 92;
			goto inconsistent;
		}
		if (find_word(p->users,args[2]) == NULL) {
			p->users = add_word_to_list(p->users,args[2]);
		}
		break;
	case DB_REMOVE_USER :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 93;
			goto inconsistent;
		}
		wl1 = p->users;
		wl2 = NULL;
		while (wl1 != NULL) {
			if (strcmp(args[2],wl1->word) != 0)
				wl2 = add_word_to_list(wl2,wl1->word);
			wl1 = wl1->next;
		}
		clear_word_list(p->users);
		p->users = wl2;
		break;
	case DB_SET_NEXT_ENTRY :
		p = find_prodsys(*ps,args[1]);
		if (p == NULL) {
			incon_no = 96;
			goto inconsistent;
		}
		p->next_entry = atoi(args[2]);
		break;
	case DB_NO_MORE_UPDATES :
		return(-1);
	default:
		return(-1);
	}
	return(0);


inconsistent:
	sprintf(incon_msg,"database inconsistency #%d",incon_no);
	if (incon_no == 76)
		printerr(FACILITY,ERROR_MSG,incon_msg,
		    "User may have called acp_done_with_tape with bad device name",
		    3);
	else
		printerr(FACILITY,ERROR_MSG,incon_msg,
		    msg_build("db_update",args),4);
	return(1);

}

/******************************************************************/
/*
    FUNCTION NAME:

      build_update_msg

  DESCRIPTION:

      Builds a database update message.

  FORMAL PARAMETERS:

      int msgtyp - The type of message needed.
      union itemptrs *d - An array of pointers to items in the database.

  FUNCTION VALUE:

      (char *) - Database update message.

								  */
/******************************************************************/

char *build_update_msg(msgtyp,d)
int msgtyp;
union itemptrs d[];
{

	static char msg[256];

	char *m;

	char *t0,*t1,*t2;

	sprintf(msg,"db_update|%d",msgtyp);

	m = &msg[strlen(msg)];

	switch (msgtyp) {
	case DB_ADD_CPU :
		sprintf(m,"|%s|%d|%d|%d|%d",d[0].c->cpu_name,d[0].c->cpu_type,
		    d[0].c->status,d[0].c->power,
		    d[0].c->task);
		break;
	case DB_MODIFY_CPU :
		sprintf(m,"|%s|%d|%d|%d|%d",d[0].c->cpu_name,d[0].c->cpu_type,
		    d[0].c->status,d[0].c->power,
		    d[0].c->task);
		break;
	case DB_DELETE_CPU :
		sprintf(m,"|%s",d[0].c->cpu_name);
		break;
	case DB_ADD_TAPE :
		sprintf(m,"|%s|%d|%s|%d|%d|%s",d[0].t->dev_name,d[0].t->dev_type,
		    d[0].t->cpu->cpu_name,d[0].t->status,
		    d[0].t->used,d[0].t->physical);
		break;
	case DB_MODIFY_TAPE :
		sprintf(m,"|%s|%d|%s|%d|%d",d[0].t->dev_name,d[0].t->dev_type,
		    d[0].t->cpu->cpu_name,d[0].t->status,
		    d[0].t->used);
		break;
	case DB_DELETE_TAPE :
		sprintf(m,"|%s|%s",d[0].t->dev_name,d[0].t->cpu->cpu_name);
		break;
	case DB_ADD_FILESYSTEM :
		sprintf(m,"|%s|%s|%d|%d",d[0].f->cpu->cpu_name,
		    d[0].f->mount_point,d[0].f->blocks,d[0].f->id);
		break;
	case DB_REMOVE_FILESYSTEM :
		sprintf(m,"|%d",d[0].f->id);
		break;
	case DB_ADD_PRODSYS :
		sprintf(m,"|%s|%s|%d|%d|%s|%s|%s|%d|not_used",
		    d[0].p->ps_name,
		    d[0].p->jm_cpu->cpu_name,
		    d[0].p->status,d[0].p->next_entry,
		    d[0].p->account,d[0].p->passwd,
		    d[0].p->login_dir,d[0].p->jm_fs->id);
		break;
	case DB_MODIFY_PRODSYS :
		sprintf(m,"|%s|%s|%d|%d|%s|%s|%s|%d|not_used",
		    d[0].p->ps_name,
		    d[0].p->jm_cpu->cpu_name,
		    d[0].p->status,d[0].p->next_entry,
		    d[0].p->account,d[0].p->passwd,
		    d[0].p->login_dir,d[0].p->jm_fs->id);
		break;
	case DB_DELETE_PRODSYS :
		sprintf(m,"|%s",d[0].p->ps_name);
		break;
	case DB_ADD_CPU_TO_PROD :
		sprintf(m,"|%s|%s|%d|%d|not_used|%d",
		    d[0].p->ps_name,d[1].pcl->cpu->cpu_name,
		    d[1].pcl->max_proc,d[1].pcl->fs->id,
		    d[1].pcl->lognum);
		break;
	case DB_MODIFY_CPU_IN_PROD :
		sprintf(m,"|%s|%s|%d|%d|not_used|%d",
		    d[0].p->ps_name,d[1].pcl->cpu->cpu_name,
		    d[1].pcl->max_proc,d[1].pcl->fs->id,
		    d[1].pcl->lognum);
		break;
	case DB_DELETE_CPU_FROM_PROD :
		sprintf(m,"|%s|%s",d[0].p->ps_name,d[1].pcl->cpu->cpu_name);
		break;
	case DB_ADD_TAPE_TO_PROD :
		sprintf(m,"|%s|%s|%s",d[0].p->ps_name,
		    d[1].ptl->tape->dev_name,d[1].ptl->tape->cpu->cpu_name);
		break;
	case DB_DELETE_TAPE_FROM_PROD :
		sprintf(m,"|%s|%s|%s",d[0].p->ps_name,
		    d[1].ptl->tape->dev_name,d[1].ptl->tape->cpu->cpu_name);
		break;
	case DB_ADD_USER :
		sprintf(m,"|%s|%s",
		    d[0].p->ps_name,d[1].str);
		break;
	case DB_REMOVE_USER :
		sprintf(m,"|%s|%s",
		    d[0].p->ps_name,d[1].str);
		break;
	case DB_ALLOCATE_AREA :
		sprintf(m,"|%s|%d",d[0].p->ps_name,d[1].i);
		break;
	case DB_FREE_AREA :
		sprintf(m,"|%s|%d",d[0].p->ps_name,d[1].i);
		break;
	case DB_ADD_JOB :
		sprintf(m,"|%s|%d|%d|%d|%d",
		    d[0].p->ps_name,d[1].pjl->entry,
		    d[1].pjl->pipestage,d[1].pjl->status,
		    d[1].pjl->area);
		break;
	case DB_FILL_SUBMIT_INFO :
		sprintf(m,"|%s|%s|%s|%s|%s|%s|%s",
		    d[0].p->ps_name,d[1].pjl->sub->jcf,
		    d[1].pjl->sub->jdf,
		    d[1].pjl->sub->username,d[1].pjl->sub->node,
		    d[1].pjl->sub->time,d[1].pjl->sub->address);
		break;
	case DB_ADVANCE_JOB :
		sprintf(m,"|%s|%d|%s|%d",
		    d[0].p->ps_name,d[1].pjl->pipestage,d[2].str,d[1].pjl->entry);
		break;
	case DB_REMOVE_JOB :
		sprintf(m,"|%s|%d|%d|%d",d[0].p->ps_name,d[1].pjl->pipestage
		    ,d[1].pjl->entry,d[1].pjl->abort);
		break;
	case DB_UPDATE_JOB_STATUS :
		sprintf(m,"|%s|%d|%d|%d",d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[1].pjl->status);
		break;
	case DB_SET_JOB_AREA :
		sprintf(m,"|%s|%d|%d|%d",d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].i);
		break;
	case DB_ADD_JOB_CLASS :
		sprintf(m,"|%s|%d",d[0].p->ps_name,d[2].cl->class);
		break;
	case DB_ADD_TAPE_LIST :
		sprintf(m,"|%s|%d|%d|%d|%d",d[0].p->ps_name,d[2].cl->class,
		    d[3].tlh->listnum,d[3].tlh->io,d[3].tlh->devtype);
		break;
	case DB_ADD_TAPE_REEL :
		sprintf(m,"|%s|%d|%d|%d|%d|%s",d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].cl->class,
		    d[3].tlh->listnum,d[4].otl->vsn);
		break;
	case DB_INSERT_TAPE_REEL :
		sprintf(m,"|%s|%d|%d|%d|%d|%s|%s",
		    d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].cl->class,
		    d[3].tlh->listnum,d[4].str,d[5].str);
		break;
	case DB_REMOVE_TAPE_REEL :
		sprintf(m,"|%s|%d|%d|%d|%d|%s",d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].cl->class,
		    d[3].tlh->listnum,d[4].otl->vsn);
		break;
	case DB_TAPE_REEL_STATUS :
		sprintf(m,"|%s|%d|%d|%d|%d|%s|%d",
		    d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].cl->class,
		    d[3].tlh->listnum,d[4].otl->vsn,d[4].otl->status);
		break;
	case DB_TAPE_TIME_REQ :
		sprintf(m,"|%s|%d|%d|%d|%d|%s|%ld",
		    d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].cl->class,
		    d[3].tlh->listnum,d[4].otl->vsn,d[4].otl->time_req);
		break;
	case DB_TAPE_TIME_MNT :
		sprintf(m,"|%s|%d|%d|%d|%d|%s|%ld",
		    d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].cl->class,
		    d[3].tlh->listnum,d[4].otl->vsn,d[4].otl->time_mnt);
		break;
	case DB_TAPE_TIME_DONE :
		sprintf(m,"|%s|%d|%d|%d|%d|%s|%ld",
		    d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[2].cl->class,
		    d[3].tlh->listnum,d[4].otl->vsn,d[4].otl->time_done);
		break;
	case DB_MOVE_JOB_TO_QUEUE :
		sprintf(m,"|%s",d[0].p->ps_name);
		break;
	case DB_MOVE_JOB_TO_ON_DECK :
		sprintf(m,"|%s",d[0].p->ps_name);
		break;
	case DB_MOVE_JOB_TO_CURRENT :
		sprintf(m,"|%s",d[0].p->ps_name);
		break;
	case DB_MOVE_JOB_TO_RUNDOWN :
		sprintf(m,"|%s",d[0].p->ps_name);
		break;
	case DB_SET_ABORT_FLAG :
		sprintf(m,"|%s|%d|%d",d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry);
		break;
	case DB_CLEAR_ABORT_FLAG :
		sprintf(m,"|%s|%d|%d",d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry);
		break;
	case DB_JCSS_PID :
		sprintf(m,"|%s|%d|%d|%d",d[0].p->ps_name,d[1].pjl->pipestage,
		    d[1].pjl->entry,d[1].pjl->jcss_pid);
		break;
	case DB_SET_PHASE_TIMES :
		if (d[1].pjl->on_deck_time == NULL)
			t0 = "null";
		else
			t0 = d[1].pjl->on_deck_time;
		if (d[1].pjl->start_time == NULL)
			t1 = "null";
		else
			t1 = d[1].pjl->start_time;
		if (d[1].pjl->rundown_time == NULL)
			t2 = "null";
		else
			t2 = d[1].pjl->rundown_time;
		sprintf(m,"|%s|%s|%s|%s",d[0].p->ps_name,t0,t1,t2);
		break;
	case DB_MOUNT_INFO :
		sprintf(m,"|%s|%s|%s|%d|%d",
		    d[0].t->cpu->cpu_name,d[0].t->dev_name,
		    d[0].t->mi->vsn,d[0].t->mi->rstate,
		    d[0].t->mi->mstate);
		break;
	case DB_SET_NEXT_ENTRY :
		sprintf(m,"|%s|%d",d[0].p->ps_name,d[1].i);
		break;
	case DB_NO_MORE_UPDATES :
		break;
	default:
		return(NULL);
	}

	return(msg);
}

/******************************************************************/
/*
    FUNCTION NAME:

      log_database_update

								  */
/******************************************************************/

int log_database_update(args,fp)
char **args;
FILE *fp;
{
	int msgtyp;

	msgtyp = atoi(args[0]);

	switch (msgtyp) {
	case DB_ADD_CPU :
		fprintf(fp,"add cpu; ");
		fprintf(fp,"%s, type=%s, status=%s\n",args[1],
		    cpu_types[atoi(args[2])],hardware_state[atoi(args[3])]);
		break;
	case DB_MODIFY_CPU :
		fprintf(fp,"modify cpu; ");
		fprintf(fp,"%s, type=%s, status=%s\n",args[1],
		    cpu_types[atoi(args[2])],hardware_state[atoi(args[3])]);
		break;
	case DB_DELETE_CPU :
		fprintf(fp,"delete cpu; %s\n",args[1]);
		break;
	case DB_ADD_TAPE :
		fprintf(fp,"add tape; ");
		fprintf(fp,"%s, type=%s, cpu=%s, status=%s, alloc=%s, phys=%s\n",
		    args[1],device_types[atoi(args[2])],
		    args[3],hardware_state[atoi(args[4])],
		    allocation_status[atoi(args[5])],args[6]);
		break;
	case DB_MODIFY_TAPE :
		fprintf(fp,"modify tape; ");
		fprintf(fp,"%s, type=%s, cpu=%s, status=%s, alloc=%s\n",
		    args[1],device_types[atoi(args[2])],
		    args[3],hardware_state[atoi(args[4])],
		    allocation_status[atoi(args[5])]);
		break;
	case DB_DELETE_TAPE :
		fprintf(fp,"delete tape; ");
		fprintf(fp,"%s, cpu=%s\n",args[1],args[2]);
		break;
	case DB_ADD_FILESYSTEM :
		fprintf(fp,"add filesystem; cpu=%s, mp=%s, id=%s\n",
		    args[1],args[2],args[4]);
		break;
	case DB_REMOVE_FILESYSTEM :
		fprintf(fp,"remove filesystem; id=%s\n",args[1]);
		break;
	case DB_ADD_PRODSYS :
		fprintf(fp,"add prodsys; ");
		fprintf(fp,"%s, jm_cpu=%s, next_entry=%s\n",
		    args[1],args[2],args[4]);
		break;
	case DB_MODIFY_PRODSYS :
		fprintf(fp,"modify prodsys; ");
		fprintf(fp,"%s, jm_cpu=%s, next_entry=%s\n",
		    args[1],args[2],args[4]);
		break;
	case DB_DELETE_PRODSYS :
		fprintf(fp,"delete prodsys; %s\n",args[1]);
		break;
	case DB_ADD_CPU_TO_PROD :
		fprintf(fp,"add cpu to prodsys; ");
		fprintf(fp,"%s, cpu=%s, max_proc=%s\n",args[1],args[2],args[3]);
		break;
	case DB_MODIFY_CPU_IN_PROD :
		fprintf(fp,"mod cpu in prodsys; ");
		fprintf(fp,"%s, cpu=%s, max_proc=%s\n",args[1],args[2],args[3]);
		break;
	case DB_DELETE_CPU_FROM_PROD :
		fprintf(fp,"del cpu from prodsys; %s, cpu=%s\n",
		    args[1],args[2]);
		break;
	case DB_ADD_TAPE_TO_PROD :
		fprintf(fp,"add tape to prodsys; ");
		fprintf(fp,"%s, dev=%s, cpu=%s\n",args[1],args[2],args[3]);
		break;
	case DB_DELETE_TAPE_FROM_PROD :
		fprintf(fp,"del tape from prodsys; ");
		fprintf(fp,"%s, dev=%s, cpu=%s\n",args[1],args[2],args[3]);
		break;
	case DB_ADD_USER :
		fprintf(fp,"add user; %s, user=%s\n",args[1],args[2]);
		break;
	case DB_REMOVE_USER :
		fprintf(fp,"remove user; %s, user=%s\n",args[1],args[2]);
		break;
	case DB_ALLOCATE_AREA :
		fprintf(fp,"allocate area; %s, area=%s\n",args[1],args[2]);
		break;
	case DB_FREE_AREA :
		fprintf(fp,"free area; %s, area=%s\n",args[1],args[2]);
		break;
	case DB_ADD_JOB :
		fprintf(fp,"add job; ");
		fprintf(fp,"%s, entry=%s, pipestage=%s, status=%s, area=%s\n",
		    args[1],args[2],job_stages[atoi(args[3])],
		    job_statuses[atoi(args[4])],args[5]);
		break;
	case DB_FILL_SUBMIT_INFO :
		fprintf(fp,"submit info; ");
		fprintf(fp,"%s, jcf=%s, jdf=%s, user=%s, node=%s, time=%s\n",
		    args[1],args[2],args[3],args[4],args[5],args[6]);
		break;
	case DB_ADVANCE_JOB :
		fprintf(fp,"advance job; ");
		fprintf(fp,"%s, stage=%s, time=%s, entry=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[4]);
		break;
	case DB_REMOVE_JOB :
		fprintf(fp,"remove job; ");
		fprintf(fp,"%s, stage=%s, entry=%s, abort=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[4]);
		break;
	case DB_UPDATE_JOB_STATUS :
		fprintf(fp,"job status; ");
		fprintf(fp,"%s, stage=%s, entry=%s, status=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],
		    job_statuses[atoi(args[4])]);
		break;
	case DB_SET_JOB_AREA :
		fprintf(fp,"set job area; ");
		fprintf(fp,"%s, stage=%s, entry=%s, area=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[4]);
		break;
	case DB_ADD_JOB_CLASS :
		fprintf(fp,"add job class; %s, class=%s\n",args[1],args[2]);
		break;
	case DB_ADD_TAPE_LIST :
		fprintf(fp,"add tape list; ");
		fprintf(fp,"%s, class=%s, list=%s, direction=%s, devtype=%s\n",
		    args[1],args[2],args[3],tape_direction[atoi(args[4])],
		    reqdev_types[atoi(args[5])]);
		break;
	case DB_ADD_TAPE_REEL :
		fprintf(fp,"add tape reel; %s\n",args[1]);
		/*
	    sprintf(m,"|%s|%d|%d|%d|%d|%s",d[0].p->ps_name,d[1].pjl->pipestage,
		d[1].pjl->entry,d[2].cl->class,
		d[3].tlh->listnum,d[4].otl->vsn);
*/
		break;
	case DB_INSERT_TAPE_REEL :
		fprintf(fp,"insert tape reel; %s\n",args[1]);
		/*
	    sprintf(m,"|%s|%d|%d|%d|%d|%s|%s",
		d[0].p->ps_name,d[1].pjl->pipestage,
		d[1].pjl->entry,d[2].cl->class,
		d[3].tlh->listnum,d[4].str,d[5].str);
*/
		break;
	case DB_REMOVE_TAPE_REEL :
		fprintf(fp,"remove tape reel; %s\n",args[1]);
		/*
	    sprintf(m,"|%s|%d|%d|%d|%d|%s",d[0].p->ps_name,d[1].pjl->pipestage,
		d[1].pjl->entry,d[2].cl->class,
		d[3].tlh->listnum,d[4].otl->vsn);
*/
		break;
	case DB_TAPE_REEL_STATUS :
		fprintf(fp,"tape reel status; ");
		fprintf(fp,"%s, stage=%s, entry=%s, vsn=%s, status=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[6],
		    tape_reel_status[atoi(args[7])]);
		break;
	case DB_TAPE_TIME_REQ :
		fprintf(fp,"tape time req; ");
		fprintf(fp,"%s, stage=%s, entry=%s, vsn=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[6]);
		break;
	case DB_TAPE_TIME_MNT :
		fprintf(fp,"tape time mnt; ");
		fprintf(fp,"%s, stage=%s, entry=%s, vsn=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[6]);
		break;
	case DB_TAPE_TIME_DONE :
		fprintf(fp,"tape time done; ");
		fprintf(fp,"%s, stage=%s, entry=%s, vsn=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[6]);
		break;
	case DB_MOVE_JOB_TO_QUEUE :
		fprintf(fp,"move job to queue; %s\n",args[1]);
		break;
	case DB_MOVE_JOB_TO_ON_DECK :
		fprintf(fp,"move job to on_deck; %s\n",args[1]);
		break;
	case DB_MOVE_JOB_TO_CURRENT :
		fprintf(fp,"move job to current; %s\n",args[1]);
		break;
	case DB_MOVE_JOB_TO_RUNDOWN :
		fprintf(fp,"move job to rundown; %s\n",args[1]);
		break;
	case DB_SET_ABORT_FLAG :
		fprintf(fp,"set abort flag; ");
		fprintf(fp,"%s, stage=%s, entry=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3]);
		break;
	case DB_CLEAR_ABORT_FLAG :
		fprintf(fp,"clear abort flag; ");
		fprintf(fp,"%s, stage=%s, entry=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3]);
		break;
	case DB_JCSS_PID :
		fprintf(fp,"jcss pid; ");
		fprintf(fp,"%s, stage=%s, entry=%s, jcss_pid=%s\n",
		    args[1],job_stages[atoi(args[2])],args[3],args[4]);
		break;
	case DB_SET_PHASE_TIMES :
		fprintf(fp,"set phase times; %s",args[1]);
		break;
	case DB_MOUNT_INFO :
		fprintf(fp,"mount info; ");
		fprintf(fp,"cpu=%s, dev=%s, vsn=%s, rstate=%s, mstage=%s\n",
		    args[1],args[2],args[3],args[4],mount_state[atoi(args[5])]);
		break;
	case DB_SET_NEXT_ENTRY :
		fprintf(fp,"set next entry; %s, next_entry=%s\n",args[1],args[2]);
		break;
	case DB_NO_MORE_UPDATES :
		fprintf(fp,"no more updates\n");
		break;
	default:
		fprintf(fp,"db_update:\tmessage type = %d\n",msgtyp);
	}
	fflush(fp);
	return 0;
}

/******************************************************************/
/*
    FUNCTION NAME:

      build_query_msg

  DESCRIPTION:

      Builds a database query message.

  FORMAL PARAMETERS:

      int msgtyp - The type of message needed.
      char *s1,*s2 - Parameters describing the item.

  FUNCTION VALUE:

      (char *) - Database query message.

								  */
/******************************************************************/

char *build_query_msg(msgtyp,s1,s2)
int msgtyp;
char *s1,*s2;
{

	static char msg[256];

	char *m;

	sprintf(msg,"db_query|%d",msgtyp);

	m = &msg[strlen(msg)];

	switch (msgtyp) {
	case DB_QUERY_CPU :
		sprintf(m,"|%s",s1);
		break;
	case DB_QUERY_TAPE :
		sprintf(m,"|%s|%s",s1,s2);
		break;
	case DB_QUERY_FS :
		sprintf(m,"|%s",s1);
		break;
	case DB_QUERY_DB :
		break;
	case DB_QUERY_RESOURCES :
		break;
	case DB_QUERY_CPU_LIST :
		break;
	case DB_QUERY_TAPE_LIST :
		break;
	case DB_QUERY_FS_LIST :
		break;
	case DB_QUERY_PROD_CPU_LIST :
		sprintf(m,"|%s",s1);
		break;
	case DB_QUERY_PROD_TAPE_LIST :
		sprintf(m,"|%s",s1);
		break;
	case DB_QUERY_PROD_DEF :
		sprintf(m,"|%s",s1);
		break;
	case DB_QUERY_JOB :
		sprintf(m,"|%s|%s",s1,s2);
		break;
	case DB_QUERY_PROD_RESOURCES :
		sprintf(m,"|%s",s1);
		break;
	case DB_QUERY_JOB_INFO :
		sprintf(m,"|%s",s1);
		break;
	default:
		return(NULL);
	}

	return(msg);
}

/******************************************************************/
/*
    FUNCTION NAME:

      process_query_msg

  DESCRIPTION:

      Takes a database query message and generates database update messages
      ending with a DB_NO_MORE_UPDATES message.

  FORMAL PARAMETERS:

      struct fs_list **fs - The first entry point into the database -
			        the pointer to the list of filesystems.

      struct cpu_list **cpu - The second entry point into the database -
			      the pointer to the list of CPUs.

      struct tape_drive_list **tape - The third entry point into the database -
				      the pointer to the list of tapes.

      struct prod_sys **ps - The last entry point into the database -
			     the pointer to the list of production systems.

      char *msg - Query message.

      int (*qf)() - Pointer to a function which will process each database
		    update message generated.

  SIDE EFFECTS:

      Parses the message causing the external variables msg_cmd and msg_args
      to be set to new values.

								  */
/******************************************************************/

process_query_msg(fs,cpu,tape,ps,msg,qf)
struct fs_list **fs;
struct cpu_list **cpu;
struct tape_drive_list **tape;
struct prod_sys **ps;
char *msg;
int (*qf)();
{
	union itemptrs d[6],x[6];

	struct fs_list *f;
	struct cpu_list *c;
	struct tape_drive_list *t;
	struct prod_sys *p;
	struct prod_tape_list *pt;
	struct prod_cpu_list *pc;
	struct class_list *cl;
	struct tape_list_head *tlh;
	struct ordered_tape_list *otl;
	struct word_list *wl;

	int msgtyp;

	int pipestage;

	static int top = 0;

	char *update_msg,*query_msg;

	int incon_no;
	char incon_msg[80];

	if (top == 0) {
		top = -1;
		process_query_msg(fs,cpu,tape,ps,msg,qf);
		(*qf)(build_update_msg(DB_NO_MORE_UPDATES,d));
		top = 0;
		return;
	}

	msg_parse(msg);

	if (strcmp(msg_cmd,"db_query") != 0) {
		printerr(FACILITY,ERROR_MSG,"not a query message, cannot process",
		    NULL,5);
		return;
	}

	msgtyp = atoi(msg_args[0]);
	switch (msgtyp) {
	case DB_QUERY_CPU :
		d[0].c = find_cpu(*cpu,msg_args[1]);
		if (d[0].c != NULL) {
			(*qf)(build_update_msg(DB_ADD_CPU,d));
		}
		else {
			incon_no = 1;
			goto inconsistent;
		}
		break;
	case DB_QUERY_TAPE :
		d[0].t = find_tape(*tape,msg_args[1],msg_args[2]);
		if (d[0].t != NULL) {
			(*qf)(build_update_msg(DB_ADD_TAPE,d));
			(*qf)(build_update_msg(DB_MOUNT_INFO,d));
		}
		else {
			incon_no = 2;
			goto inconsistent;
		}
		break;
	case DB_QUERY_FS :
		d[0].f = find_fs(*fs,atoi(msg_args[1]));
		if (d[0].f != NULL) {
			(*qf)(build_update_msg(DB_ADD_FILESYSTEM,d));
		}
		else {
			incon_no = 3;
			goto inconsistent;
		}
		break;
	case DB_QUERY_CPU_LIST :
		d[0].c = *cpu;
		while (d[0].c != NULL) {
			(*qf)(build_update_msg(DB_ADD_CPU,d));
			d[0].c = d[0].c->next;
		}
		break;
	case DB_QUERY_TAPE_LIST :
		d[0].t = *tape;
		while (d[0].t != NULL) {
			(*qf)(build_update_msg(DB_ADD_TAPE,d));
			(*qf)(build_update_msg(DB_MOUNT_INFO,d));
			d[0].t = d[0].t->next;
		}
		break;
	case DB_QUERY_FS_LIST :
		d[0].f = *fs;
		while (d[0].f != NULL) {
			(*qf)(build_update_msg(DB_ADD_FILESYSTEM,d));
			d[0].f = d[0].f->next;
		}
		break;
	case DB_QUERY_PROD_CPU_LIST :
		d[0].p = find_prodsys(*ps,msg_args[1]);
		if (d[0].p != NULL) {
			d[1].pcl = d[0].p->cpus;
			while (d[1].pcl != NULL) {
				(*qf)(build_update_msg(DB_ADD_CPU_TO_PROD,d));
				d[1].pcl = d[1].pcl->next;
			}
		}
		else {
			incon_no = 4;
			goto inconsistent;
		}
		break;
	case DB_QUERY_PROD_TAPE_LIST :
		d[0].p = find_prodsys(*ps,msg_args[1]);
		if (d[0].p != NULL) {
			d[1].ptl = d[0].p->tapes;
			while (d[1].ptl != NULL) {
				(*qf)(build_update_msg(DB_ADD_TAPE_TO_PROD,d));
				d[1].ptl = d[1].ptl->next;
			}
		}
		else {
			incon_no = 5;
			goto inconsistent;
		}
		break;
	case DB_QUERY_PROD_DEF :
		d[0].p = find_prodsys(*ps,msg_args[1]);
		if (d[0].p != NULL) {
			(*qf)(build_update_msg(DB_ADD_PRODSYS,d));
			x[0].p = d[0].p;
			if (d[0].p->job_area[0] == -1) {
				x[1].i = 1;
				(*qf)(build_update_msg(DB_ALLOCATE_AREA,x));
			}
			if (d[0].p->job_area[1] == -1) {
				x[1].i = 2;
				(*qf)(build_update_msg(DB_ALLOCATE_AREA,x));
			}
			if (d[0].p->job_area[2] == -1) {
				x[1].i = 3;
				(*qf)(build_update_msg(DB_ALLOCATE_AREA,x));
			}
			query_msg = build_query_msg(DB_QUERY_PROD_CPU_LIST,
			    d[0].p->ps_name,NULL);
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);
			query_msg = build_query_msg(DB_QUERY_PROD_TAPE_LIST,
			    d[0].p->ps_name,NULL);
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);

			wl = d[0].p->users;
			while (wl != NULL) {
				x[1].str = wl->word;
				(*qf)(build_update_msg(DB_ADD_USER,x));
				wl = wl->next;
			}
		}
		else {
			incon_no = 6;
			goto inconsistent;
		}
		break;
	case DB_QUERY_DB :
		query_msg = build_query_msg(DB_QUERY_CPU_LIST,NULL,NULL);
		process_query_msg(fs,cpu,tape,ps,query_msg,qf);
		query_msg = build_query_msg(DB_QUERY_TAPE_LIST,NULL,NULL);
		process_query_msg(fs,cpu,tape,ps,query_msg,qf);
		query_msg = build_query_msg(DB_QUERY_FS_LIST,NULL,NULL);
		process_query_msg(fs,cpu,tape,ps,query_msg,qf);
		p = *ps;
		while (p != NULL) {
			query_msg = build_query_msg(DB_QUERY_PROD_DEF,p->ps_name,NULL);
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);
			query_msg = build_query_msg(DB_QUERY_JOB_INFO,p->ps_name,NULL);
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);
			p = p->next;
		}
		break;
	case DB_QUERY_RESOURCES :
		query_msg = build_query_msg(DB_QUERY_CPU_LIST,NULL,NULL);
		process_query_msg(fs,cpu,tape,ps,query_msg,qf);
		query_msg = build_query_msg(DB_QUERY_TAPE_LIST,NULL,NULL);
		process_query_msg(fs,cpu,tape,ps,query_msg,qf);
		query_msg = build_query_msg(DB_QUERY_FS_LIST,NULL,NULL);
		process_query_msg(fs,cpu,tape,ps,query_msg,qf);
		p = *ps;
		while (p != NULL) {
			query_msg = build_query_msg(DB_QUERY_PROD_DEF,p->ps_name,NULL);
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);
			p = p->next;
		}
		break;
	case DB_QUERY_JOB :
		d[0].p = find_prodsys(*ps,msg_args[1]);
		if (d[0].p == NULL) {
			incon_no = 7;
			goto inconsistent;
		}

		if (msg_args[2][0] == 'n') {
			d[1].pjl = d[0].p->newjob;
			if (d[1].pjl == NULL) {
				incon_no = 8;
				goto inconsistent;
			}
		}
		else if (msg_args[2][0] == 'q') {
			int entnum;
			struct prod_job_list *pjl;
			entnum = atoi(&msg_args[2][1]);
			d[1].pjl = d[0].p->job_queue;
			while (d[1].pjl != NULL) {
				if (entnum == d[1].pjl->entry) break;
				d[1].pjl = d[1].pjl->next;
			}
			if (d[1].pjl == NULL) {		/* Job not found in queue */
				incon_no = 9;
				goto inconsistent;
			}
		}
		else if (msg_args[2][0] == 'o') {
			d[1].pjl = d[0].p->on_deck;
			if (d[1].pjl == NULL) {
				incon_no = 10;
				goto inconsistent;
			}
		}
		else if (msg_args[2][0] == 'c') {
			d[1].pjl = d[0].p->current;
			if (d[1].pjl == NULL) {
				incon_no = 11;
				goto inconsistent;
			}
		}
		else if (msg_args[2][0] == 'r') {
			d[1].pjl = d[0].p->rundown;
			if (d[1].pjl == NULL) {
				incon_no = 12;
				goto inconsistent;
			}
		}

		(*qf)(build_update_msg(DB_ADD_JOB,d));
		(*qf)(build_update_msg(DB_FILL_SUBMIT_INFO,d));
		(*qf)(build_update_msg(DB_SET_PHASE_TIMES,d));

		d[2].cl = d[1].pjl->clist;
		while (d[2].cl != NULL) {
			(*qf)(build_update_msg(DB_ADD_JOB_CLASS,d));
			d[3].tlh = d[2].cl->tlh;
			while (d[3].tlh != NULL) {
				(*qf)(build_update_msg(DB_ADD_TAPE_LIST,d));
				d[4].otl = d[3].tlh->tlist;
				pipestage = d[1].pjl->pipestage;
				d[1].pjl->pipestage = DB_PIPE_NEW;
				while (d[4].otl != NULL) {
					(*qf)(build_update_msg(DB_ADD_TAPE_REEL,d));
					(*qf)(build_update_msg(DB_TAPE_REEL_STATUS,d));
					if (d[4].otl->time_req != 0)
						(*qf)(build_update_msg(DB_TAPE_TIME_REQ,d));
					if (d[4].otl->time_mnt != 0)
						(*qf)(build_update_msg(DB_TAPE_TIME_MNT,d));
					if (d[4].otl->time_done != 0)
						(*qf)(build_update_msg(DB_TAPE_TIME_DONE,d));
					d[4].otl = d[4].otl->next;
				}
				d[1].pjl->pipestage = pipestage;
				d[3].tlh = d[3].tlh->next;
			}
			d[2].cl = d[2].cl->next;
		}
		break;
	case DB_QUERY_PROD_RESOURCES :
		p = find_prodsys(*ps,msg_args[1]);
		if (p == NULL) {
			incon_no = 13;
			goto inconsistent;
		}

		pc = p->cpus;		/* Clear each filesystem's mark */
		while (pc != NULL) {
			pc->fs->mark = 0;
			pc = pc->next;
		}

		d[0].c = p->jm_cpu;		/* JM CPU and filesystem */
		(*qf)(build_update_msg(DB_ADD_CPU,d));
		if (p->jm_cpu != p->jm_fs->cpu) {
			d[0].c = p->jm_fs->cpu;
			(*qf)(build_update_msg(DB_ADD_CPU,d));
		}
		d[0].f = p->jm_fs;
		(*qf)(build_update_msg(DB_ADD_FILESYSTEM,d));
		p->jm_fs->mark = -1;

		pc = p->cpus;		/* CPU list */
		while (pc != NULL) {
			d[0].c = pc->cpu;
			(*qf)(build_update_msg(DB_ADD_CPU,d));
			if (pc->fs->mark == 0) {
				d[0].c = pc->fs->cpu;
				(*qf)(build_update_msg(DB_ADD_CPU,d));
				d[0].f = pc->fs;
				(*qf)(build_update_msg(DB_ADD_FILESYSTEM,d));
				pc->fs->mark = -1;
			}
			pc = pc->next;
		}

		pt = p->tapes;		/* Tape list */
		while (pt != NULL) {
			d[0].t = pt->tape;
			(*qf)(build_update_msg(DB_ADD_TAPE,d));
			(*qf)(build_update_msg(DB_MOUNT_INFO,d));
			pt = pt->next;
		}
		break;
	case DB_QUERY_JOB_INFO :

		p = find_prodsys(*ps,msg_args[1]);
		if (p == NULL) {
			incon_no = 14;
			goto inconsistent;
		}
		d[0].p = p;

		if (p->rundown != NULL) {
			query_msg = build_query_msg(DB_QUERY_JOB,p->ps_name,"r");
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);
			(*qf)(build_update_msg(DB_MOVE_JOB_TO_RUNDOWN,d));
			d[1].pjl = p->rundown;
			if (d[1].pjl->abort == -1)	/* Set the abort flag if necessary */
				(*qf)(build_update_msg(DB_SET_ABORT_FLAG,d));
			(*qf)(build_update_msg(DB_JCSS_PID,d));
		}

		if (p->current != NULL) {
			query_msg = build_query_msg(DB_QUERY_JOB,p->ps_name,"c");
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);
			(*qf)(build_update_msg(DB_MOVE_JOB_TO_CURRENT,d));
			d[1].pjl = p->current;
			if (d[1].pjl->abort == -1)	/* Set the abort flag if necessary */
				(*qf)(build_update_msg(DB_SET_ABORT_FLAG,d));
			(*qf)(build_update_msg(DB_JCSS_PID,d));
		}

		if (p->on_deck != NULL) {
			query_msg = build_query_msg(DB_QUERY_JOB,p->ps_name,"o");
			process_query_msg(fs,cpu,tape,ps,query_msg,qf);
			(*qf)(build_update_msg(DB_MOVE_JOB_TO_ON_DECK,d));
			d[1].pjl = p->on_deck;
			if (d[1].pjl->abort == -1)	/* Set the abort flag if necessary */
				(*qf)(build_update_msg(DB_SET_ABORT_FLAG,d));
			(*qf)(build_update_msg(DB_JCSS_PID,d));
		}

		{
			struct prod_job_list *q;
			char a[6];
			sort_job_queue(p);
			q = p->job_queue;
			while (q != NULL) {
				sprintf(a,"q%d",q->entry);
				query_msg = build_query_msg(DB_QUERY_JOB,p->ps_name,a);
				process_query_msg(fs,cpu,tape,ps,query_msg,qf);
				(*qf)(build_update_msg(DB_MOVE_JOB_TO_QUEUE,d));
				d[1].pjl = q;
				if (d[1].pjl->abort == -1)	/* Set the abort flag if necessary */
					(*qf)(build_update_msg(DB_SET_ABORT_FLAG,d));
				(*qf)(build_update_msg(DB_JCSS_PID,d));
				q = q->next;
			}
		}
		/*
	    if (p->newjob != NULL) {
		query_msg = build_query_msg(DB_QUERY_JOB,p->ps_name,"n");
		process_query_msg(fs,cpu,tape,ps,query_msg,qf);
	    }
*/
		break;
	}

	return;

inconsistent:
	sprintf(incon_msg,"database query inconsistency #%d",incon_no);
	printerr(FACILITY,ERROR_MSG,incon_msg,NULL,6);
	return;
}

/******************************************************************/
/*
    FUNCTION NAME:

      load_update_messages

  DESCRIPTION:

      Reads a file containing database update messages and builds a database.

  FORMAL PARAMETERS:

      struct fs_list **fs - The first entry point into the database -
			        the pointer to the list of filesystems.

      struct cpu_list **cpu - The second entry point into the database -
			      the pointer to the list of CPUs.

      struct tape_drive_list **tape - The third entry point into the database -
				      the pointer to the list of tapes.

      struct prod_sys **ps - The last entry point into the database -
			     the pointer to the list of production systems.

      char *filename - File to open and read.

  FUNCTION VALUE:

      (int) - 0 = The file was successfully opened and processed.
	     -1 = There were errors reading the file.

								  */
/******************************************************************/

int load_update_messages(fs,cpu,tape,ps,filename)
struct fs_list **fs;
struct cpu_list **cpu;
struct tape_drive_list **tape;
struct prod_sys **ps;
char *filename;
{
	FILE *fp;

	char msg[256];
	int n;

	fp = fopen(filename,"r");
	if (fp == NULL) return(-1);
	while (fgets(msg,sizeof(msg),fp) != NULL) {
		n = strlen(msg);
		if (n > 0) msg[n-1] = '\0';
		msg_parse(msg);
		database_update(fs,cpu,tape,ps,msg_args);
	}
	fclose(fp);
	return(0);
}
