
#if 0

/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: signal_bsd.c

    Explanation: BSD signal manipulation routines.

    Exported Routines:


*/

#define _BSD_SIGNALS 1

#include <stdio.h>
#include <signal.h>
#include "../inc/perr.h"

#define FACILITY "SIGNAL"

#define MAXSTACK 256

static int stack[MAXSTACK];
static int sp = 0;

struct sigvec {
	int  (*sv_handler)();
	int  sv_mask;
	int  sv_flags;
};



/************************************************************************/
/*

    Function: sig_establish_handler

    Description: Establishes a signal handling routine for a signal.

    Arguments:

    Function Value: -1 = failed, 0 = success

*/
/************************************************************************/

int sig_establish_handler(sig,handler)
int sig;
int (*handler)();
{
	struct sigvec sv;

	sv.sv_handler = handler;
	sv.sv_mask = sigmask(sig);
	sv.sv_flags = 0;
	return sigvec(sig,&sv,0);
}

/************************************************************************/
/*

    Function: sig_block_one

    Description:

    Arguments:

    Function Value: 0 = success, -1 = failure.

*/
/************************************************************************/

int sig_block_one(sig)
int sig;
{

	/* Block the signal & store the old mask on the stack */

	stack[sp] = sigblock(sigmask(sig));
#ifdef DEBUG
	printf("sp=%d, mask=%x\n",sp,stack[sp]);
	fflush(stdout);
#endif

	/* Store the old signal mask on the stack */

	sp++;
	if (sp == MAXSTACK) {
		printerr(FACILITY,ERROR_MSG,"Signal stack overflow",NULL,1);
		return -1;
	}

	return 0;
}

/************************************************************************/
/*

    Function: sig_block_all

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int sig_block_all()
{

	/* Block the signal & store the old signal mask on the stack. */

	stack[sp] = sigblock(-1);
#ifdef DEBUG
	printf("sp=%d, mask=%x\n",sp,stack[sp]);
	fflush(stdout);
#endif

	/* Store the old signal mask on the stack */

	sp++;
	if (sp == MAXSTACK) {
		printerr(FACILITY,ERROR_MSG,"Signal stack overflow",NULL,2);
		return -1;
	}

	return 0;
}

/************************************************************************/
/*

    Function: sig_pop_mask

    Description: Pops the previous signal mask off the stack.

    Arguments:

    Function Value: 0 = success, -1 = failure.

*/
/************************************************************************/

int sig_pop_mask()
{
	if (sp == 0) {
		printerr(FACILITY,ERROR_MSG,"Signal stack underflow",
		    "A call to release signals does not have a matching call that holds signals",3);
		return -1;
	}

	sp--;
	if (sigsetmask(stack[sp]) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,4);
		return -1;
	}
#ifdef DEBUG
	printf("sp=%d, mask=%x\n",sp,stack[sp]);
	fflush(stdout);
#endif

	return 0;
}

/************************************************************************/
/*

    Function: sig_wait_ge

    Description: Wait for an integer to become greater than or equal to a
		 constant.

    Arguments:

    Function Value: -1 = failure, 0 = success.

*/
/************************************************************************/

int sig_wait_ge(sig,x,n)
int sig,*x,n;
{
	int omask;

	/* Block the signal */

	omask = sigblock(sigmask(sig));

	/* Check the condition */

check_cond:

	if (*x >= n) {
		sigsetmask(omask);
		return 0;
	}

	/* Wait until we get a signal */

	sigpause(omask);
	goto check_cond;

}

/************************************************************************/
/*

    Function: sig_wait_true

    Description: Wait for a function call to become true.

    Arguments:

    Function Value: -1 = failure, 0 = success.

*/
/************************************************************************/

int sig_wait_true(sig,func,x)
int sig,(*func)(),*x;
{
	int omask;

	/* Block the signal */

	omask = sigblock(sigmask(sig));

	/* Check the condition */

check_cond:

	if ((*func)(x)) {
		sigsetmask(omask);
		return 0;
	}

	/* Wait until we get a signal */

	sigpause(omask);
	goto check_cond;

}

#endif


#if defined(MIPS) || defined(R3000)


/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: signal_sysv.c

    Explanation: System V signal manipulation routines.

    Exported Routines:


*/

#include <stdio.h>
#include <signal.h>
#include "../inc/perr.h"

#define FACILITY "SIGNAL"

#define MAXSTACK 256

static int nholds[33] = {		 /* # of holds on signal */
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0 };

int in_sig_handler[33] = {
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0 };

static int stack[MAXSTACK];		/* Signal numbers */
static int sp = 0;


/************************************************************************/
/*

    Function: sig_establish_handler

    Description: Establishes a signal handling routine for a signal.

    Arguments:

    Function Value: -1 = failed, 0 = success

*/
/************************************************************************/

int sig_establish_handler(sig,handler)
int sig;
void (*handler)();
{
	sigset(sig,handler);
	return 0;
}

/************************************************************************/
/*

    Function: sig_block_one

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int sig_block_one(sig)
int sig;
{
	if (nholds[sig] == 0) sighold(sig);
	nholds[sig]++;
	stack[sp] = sig;
	sp++;
	if (sp == MAXSTACK) {
		printerr(FACILITY,ERROR_MSG,"Signal stack overflow",NULL,3);
		return -1;
	}
	return 0;
}

/************************************************************************/
/*

    Function: sig_block_all

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int sig_block_all()
{
	int i;

	for (i=1; i<=32; i++) {
		if (nholds[i] == 0) sighold(i);
		nholds[i]++;
	}
	stack[sp] = -1;
	sp++;

	if (sp == MAXSTACK) {
		printerr(FACILITY,ERROR_MSG,"Signal stack overflow",NULL,3);
		return -1;
	}
	return 0;
}

/************************************************************************/
/*

    Function: sig_pop_mask

    Description: Pops the previous signal mask off the stack.

    Arguments:

    Function Value: 0 = success, -1 = failure.

*/
/************************************************************************/

int sig_pop_mask()
{
	int sig,i;

	if (sp == 0) {
		printerr(FACILITY,ERROR_MSG,"Signal stack underflow",
		    "A call to release signals does not have a matching call that holds signals",6);
		return -1;
	}

	sp--;
	sig = stack[sp];
	if (sig == -1) {
		for (i=1; i<=32; i++) {
			nholds[i]--;
			if (nholds[i] == 0) {
				if (!in_sig_handler[i]) sigrelse(i);
			}
		}
	}
	else {
		nholds[sig]--;
		if (nholds[sig] == 0) {
			if (!in_sig_handler[sig]) sigrelse(sig);
		}
	}

	return 0;
}

/************************************************************************/
/*

    Function: sig_wait_ge

    Description: Wait for an integer to become greater than or equal to a
		 constant.

    Arguments:

    Function Value: -1 = failure, 0 = success.

*/
/************************************************************************/

int sig_wait_ge(sig,x,n)
int sig,*x,n;
{
	if (nholds[sig] > 0) abort();

	/* Block the signal (the signal should not be blocked already) */

	sighold(sig);

	/* Check the condition */

check_cond:

	if (*x >= n) {
		sigrelse(sig);
		return 0;	/* Success! */
	}

	/* Wait until we get a signal */

	sigpause(sig);
	goto check_cond;

}

/************************************************************************/
/*

    Function: sig_wait_true

    Description: Wait for a function call to become true.

    Arguments:

    Function Value: -1 = failure, 0 = success.

*/
/************************************************************************/

int sig_wait_true(sig,func,x)
int sig,(*func)(),*x;
{
	if (nholds[sig] > 0) abort();

	/* Hold the signal */

	sighold(sig);

	/* Check the condition */

check_cond:

	if ((*func)(x)) {
		sigrelse(sig);
		return 0;	/* Success! */
	}

	/* Wait until we get a signal */

	sigpause(sig);
	goto check_cond;

}

#else 

/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: signal.c

    Explanation: POSIX signal manipulation routines.

    Exported Routines:


*/

#include <stdio.h>
#include <signal.h>
#include "../inc/perr.h"

#define FACILITY "SIGNAL"

#define MAXSTACK 256

static sigset_t stack[MAXSTACK];
static int sp = 0;


/************************************************************************/
/*

    Function: sig_establish_handler

    Description: Establishes a signal handling routine for a signal.

    Arguments:

    Function Value: -1 = failed, 0 = success

*/
/************************************************************************/

int sig_establish_handler(sig,handler)
int sig;
void (*handler)();
{
	struct sigaction act;

	act.sa_handler = handler;
	bzero(&(act.sa_mask),sizeof(sigset_t));
	act.sa_flags = 0;

	if (sigaction(sig,&act,NULL) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,1);
		return -1;
	}
	return 0;
}

/************************************************************************/
/*

    Function: sig_block_one

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int sig_block_one(sig)
int sig;
{
	sigset_t set,oset;

	/* Create a signal mask with one bit set */

	sigemptyset(&set);
	sigaddset(&set,sig);

	/* Add the signal to the current set */

	if (sigprocmask(SIG_BLOCK,&set,&oset) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,2);
		return -1;
	}

	/* Store the old signal mask on the stack */

	stack[sp] = oset;
	sp++;
	if (sp == MAXSTACK) {
		printerr(FACILITY,ERROR_MSG,"Signal stack overflow",NULL,3);
		return -1;
	}

	return 0;
}

/************************************************************************/
/*

    Function: sig_block_all

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int sig_block_all()
{
	sigset_t set,oset;

	/* Create a signal mask with one bit set */

	sigfillset(&set);

	/* Add the signal to the current set */

	if (sigprocmask(SIG_BLOCK,&set,&oset) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,4);
		return -1;
	}

	/* Store the old signal mask on the stack */

	stack[sp] = oset;
	sp++;
	if (sp == MAXSTACK) {
		printerr(FACILITY,ERROR_MSG,"Signal stack overflow",NULL,5);
		return -1;
	}

	return 0;
}

/************************************************************************/
/*

    Function: sig_pop_mask

    Description: Pops the previous signal mask off the stack.

    Arguments:

    Function Value: 0 = success, -1 = failure.

*/
/************************************************************************/

int sig_pop_mask()
{
	if (sp == 0) {
		printerr(FACILITY,ERROR_MSG,"Signal stack underflow",
		    "A call to release signals does not have a matching call that holds signals",6);
		return -1;
	}

	sp--;
	if (sigprocmask(SIG_SETMASK,&(stack[sp]),NULL) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,7);
		return -1;
	}

	return 0;
}

/************************************************************************/
/*

    Function: sig_wait_ge

    Description: Wait for an integer to become greater than or equal to a
		 constant.

    Arguments:

    Function Value: -1 = failure, 0 = success.

*/
/************************************************************************/

int sig_wait_ge(sig,x,n)
int sig,*x,n;
{
	sigset_t set,oset;

	/* Create a signal mask with one bit set */
	sigemptyset(&set);
	sigaddset(&set,sig);

	/* Add the signal to the current set */

	if (sigprocmask(SIG_BLOCK,&set,&oset) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,8);
		return -1;
	}

	/* Check the condition */

check_cond:

	if (*x >= n) {
		if (sigprocmask(SIG_SETMASK,&oset,NULL) == -1) {
			printerr(FACILITY,ERROR_MSG,NULL,NULL,9);
			return -1;
		}
		return 0;	/* Success! */
	}

	/* Wait until we get a signal */

	sigsuspend(&oset);
	goto check_cond;

}

/************************************************************************/
/*

    Function: sig_wait_true

    Description: Wait for a function call to become true.

    Arguments:

    Function Value: -1 = failure, 0 = success.

*/
/************************************************************************/

int sig_wait_true(sig,func,x)
int sig,(*func)(),*x;
{
	sigset_t set,oset;

	/* Create a signal mask with one bit set */

	sigemptyset(&set);
	sigaddset(&set,sig);

	/* Add the signal to the current set */

	if (sigprocmask(SIG_BLOCK,&set,&oset) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,8);
		return -1;
	}

	/* Check the condition */

check_cond:

	if ((*func)(x)) {
		if (sigprocmask(SIG_SETMASK,&oset,NULL) == -1) {
			printerr(FACILITY,ERROR_MSG,NULL,NULL,9);
			return -1;
		}
		return 0;	/* Success! */
	}

	/* Wait until we get a signal */

	sigsuspend(&oset);
	goto check_cond;

}
#endif

#ifdef DEBUG

display_error(s)
char *s;
{ 
	printf("%s\n",s); 
}

main() {

	sig_block_one(SIGUSR1);
	sig_block_one(SIGINT);
	sig_block_all();
	sig_block_one(SIGHUP);
	sig_pop_mask();
	sig_pop_mask();
	sig_pop_mask();
	sig_pop_mask();
}
#endif
