#ifndef JET_HXX
#define JET_HXX

class  Jet;
struct JLterm;
struct JL;

// *******************************************************************
// Defining the environment struct

struct Jet__environment {
 double* monomial;      // Storage area for monomials used in multinomial
			//   evaluation.  The area itself is defined by 
			//   JLSetup.
 Jet*    JLmonomial;    // Storage area for JL monomials used in concatenation.
			//   The area itself is defined by JLSetup.
 int*    exponent;      // Used by nexcom (as called by JL::operator()
			//   when storing monomials. The array exponent 
			//   is allocated by JLSetup.
 char*   expCode;       // The Wilf code of a monomial. 
 char    PBOK;          // Taking Poisson brackets is OK: the dimension of 
			//   phase space is even.
 MatrixI* numPaths;     // An array counting the number of paths to a vertex 
			//   in the Wilf graph.  JLSetup arranges that 
			//   numPaths[w][n] = number of compositions of
			//   a weight  w  into  n  parts.
 int     NumVar;        // Number of scalar variables associated with 
			//   the JL variable.
 int     MaxWeight;     // Maximum acceptable weight of a JL variable,
			//   equivalent to the maximum order of derivatives
			//   to be carried through calculations.
 int     allSet;        // Records when JLSetup has been called.
 int     weightLimit;   // Used in truncated multiplication and division
 int     SpaceDim;      // This is, if you will, the dimension of phase space.
 int     dof;           // The number of degrees of freedom = SpaceDim / 2.
 double* refPoint;      // Default reference point in phase space about which 
			//   derivatives are to be taken.
 double* scale;         // An array containing numbers which scale the
                        //   problem's variables.  Each entry is a "typical
                        //   size" for the variable of the same index.
 int*    AllZeroes;     // An integer array containing zeroes.  Used to 
			//   filter the standard part of a JL variable.
 dlist   predefJL;      // A list of JL variables whose declarations precede
			//   a call to JLSetup.  These are post-processed when
			//   JLSetup is finally called.
 dlist   predefLieOp;   // A list of LieOperator variables whose declarations precede
                        //   a call to JLSetup.  These are post-processed when
                        //   JLSetup is finally called.
 Jet__environment();
 ~Jet__environment();
};


class Jet {

private:
  JL*  jl;

  Jet EpsInverse() const;  // Calculates the inverse of
                           // nilpotent Jets.

  // Friends
  friend Jet concat( const Jet& );
  friend class coord;   // ??? Is this necessary???
  friend class JetC;
  friend class LieOperator;  // ??? Necessary for LieOperator copy constructor
  friend class CLieOperator;
  friend Jet real( const JetC & );
  friend Jet imag( const JetC & );

  static dlist  coordPool;          
  static dlist  coordPtrs;          

public:
  char        stacked;    // ??? Made public for convenience: no application
                          // ??? should touch this variable.

  // Constructors and destructors_____________________________________
  Jet();
  Jet( const double );
  Jet( const Jet& );
  ~Jet();

  static void Setup( const int      = 6, /* Total number of variables */
                     const int      = 1, /* Maximum derivative weight */
                     const int      = 0, /* Dimension of phase space  */
                     const double*  = 0, /* Default reference point   */
                     const double*  = 0  /* Scaling array             */
                   );

  // Obsolete setup routines -----------------------------------------
  static void setup( const int, const int );
  static void setup( const int, const int, const double* );
  static void setup( const int, const int, const int );
  static void setup( const int, const int, const int, const double* );
  // ------------------------------------------------------------------

  static void FixReference        ( const double* );
  static void FixReference        ( const int* );
  static void FixReference        ( const Jet& );
  static void FixReferenceAtStart ( const LieOperator& );
  static void FixReferenceAtEnd   ( const LieOperator& );

  static Jet__environment oneEnv;   // ??? Make private eventually!
  static FILE  *scratchFile;

  // Public member functions__________________________________________
  JL* operator->() const { return jl; }

  JLterm* get();            // Pops the top term, which should be the 
                            // one of lowest weight.
  void addTerm( JLterm* );  // Public only for diagnostic purposes.

  void fixReference();
  void fixReference( const double* );
  void fixReference( const Jet& );
  void fixReferenceAtEnd( const LieOperator& );
  void fixReferenceAtStart( const LieOperator& );

  char isNilpotent();
  void peekAt() const;
  void printCoeffs() const;
  dlist image();
  void writeToFile( char*   /* Name of unopened file */ );
  void writeToFile( FILE* );

  void getReference( double* );
  void scaleBy( double );
  void setVariable( const double, const int );
  void setVariable( int );
  double standardPart() const;
  void clear();
  double weightedDerivative( int* );
  double derivative( int* );
  Jet filter(  const int, const int ) const;  
                                  // Returns only those JLterms whose weight 
          	                  // are between two specified values, inclusive.
  Jet filter( char (*)( const int*, const double ) ); 
                                   // Returns those JLterms for which the 
                                   // argument is satisfied.
  double operator() ( double* ) const;	   
                                   // Performs a multinomial evaluation of 
				   // the Jet variable.  Essentially acts as a 
				   // power series expansion.
  Jet  operator() ( Jet* ) const;  // Self explanatory ...
  Jet  operator() ( LieOperator& ) const; 
                                   // Self explanatory ...
  Jet  D( int* )const ;		   // Performs differentiation of a Jet variable.

  // Operators________________________________________________________
  friend ostream& operator<<(ostream&, Jet&);
  friend istream& operator>>(istream&, Jet&);
  friend char operator==( const Jet&, const Jet& );
  friend char operator==( const Jet&, const double& );
  friend char operator==( const double&, const Jet& );

  Jet& operator=( const Jet& );
  Jet& operator=( const double& );
  void operator+=( const double& );
  void operator-=( const double& );

  void operator+=( const Jet& );   // ??? Why are these void operators?

  friend Jet operator+( const Jet&, const Jet& );
  friend Jet operator+( const Jet&, const double& );
  friend Jet operator+( const double&, const Jet& );
  friend Jet operator-( const Jet&, const Jet& );
  friend Jet operator-( const Jet& );
  friend Jet operator*( const Jet&, const Jet& );
  friend Jet operator*( const Jet&, const double );
  friend Jet operator*( const double, const Jet& );  // ??? Change this!!
  friend Jet operator|( const Jet&, const Jet& );
  friend Jet operator/( const Jet&, const Jet& );
  friend Jet operator/( const Jet&, const double& ); 
  friend Jet operator/(     double, const Jet& ); 
  friend Jet operator^( const Jet&, const Jet& );   // Poisson bracket

  friend JLterm operator*( JLterm&, JLterm& );
  friend Jet operator*( const Jet&, const int );
  friend Jet operator*( const int, const Jet& );
  friend Jet operator/( const Jet&, const int& ); 

  friend LieOperator operator*( const LieOperator&, const LieOperator& );

  Jet inverse();             // Returns the inverse jet
                             //  only when the number of
                             //  coordinates = 1.

  Jet Inverse() const;       // Returns the inverse jet
                             //  only when the number of
                             //  coordinates = 1.

  // Overloaded analytic functions__________________________________
  friend Jet acos( const Jet& );
  friend Jet asin( const Jet& );
  friend Jet atan( const Jet& );
  //  friend Jet atan2( const Jet&, const Jet& );
  friend Jet cos( const Jet& );
  friend Jet cosh( const Jet& );
  friend Jet exp( const Jet& );
  friend Jet log( const Jet& );
  friend Jet log10( const Jet& );
  friend Jet pow( const Jet&, const double );
  friend Jet pow( const Jet&, int );
  friend Jet sin( const Jet& );
  friend Jet sinh( const Jet& );
  friend Jet sqrt( const Jet& );
  friend Jet tan( const Jet& );
  friend Jet tanh( const Jet& );
  friend Jet erf( const Jet& );
  friend Jet erfc( const Jet& );
  friend Jet w( const Jet& );
  friend Jet fabs( const Jet& );

  // Monomial ranking routines based on Wilf's algorithm. __________
  static void monoCode();     // Computes the code of a monomial from its
                              //   exponents.
  static void monoDecode();   // The inverse operation to monoCode.
  static int  monoRank();     // Computes the rank of a monomial.
};

void operator-=( Jet&, const Jet& );   // ??? Why are these void operators?
void operator*=( Jet&, const Jet& );   // ??? And why not member functions?
void operator/=( Jet&, const Jet& );
void operator/=( Jet&, const double );
char operator!=( const Jet&, const Jet& );
char operator!=( const Jet&, const double& );
char operator!=( const double&, const Jet& );

// *******************************************************************

class coord : public Jet {
private:
  int index;
public:
  coord();
  coord( const double );
  coord( const coord& );
  ~coord();
  coord& operator=( const coord& );
  coord& operator=( const Jet& );
  void set( const double );
} ;

#endif // Jet_HXX
