#ifndef JETC_HXX
#define JETC_HXX

struct JL;
struct JLCterm;
struct JLC;
class  Jet;
class  JetC;

// *******************************************************************
// Defining the environment struct

struct JetC__environment {
 complex* monomial;     // Storage area for monomials used in multinomial
                        //   evaluation.  The area itself is defined by
                        //   JLSetup.
 JetC*    JLmonomial;   // Storage area for JL monomials used in concatenation.
                        //   The area itself is defined by JLSetup.
 int*     exponent;     // Used by nexcom (as called by JL::operator()
                        //   when storing monomials. The array exponent
                        //   is allocated by JLSetup.
 char*    expCode;      // The Wilf code of a monomial.
 char     PBOK;         // Taking Poisson brackets is OK: the dimension of
                        //   phase space is even.
 MatrixI* numPaths;     // An array counting the number of paths to a vertex
                        //   in the Wilf graph.  JLSetup arranges that
                        //   numPaths[w][n] = number of compositions of
                        //   a weight  w  into  n  parts.
 int      NumVar;       // Number of scalar variables associated with
                        //   the JL variable.
 int      MaxWeight;    // Maximum acceptable weight of a JL variable,
                        //   equivalent to the maximum order of derivatives
                        //   to be carried through calculations.
 int      allSet;       // Records when JLSetup has been called.
 int      weightLimit;  // Used in truncated multiplication and division
 int      SpaceDim;     // This is, if you will, the dimension of phase space.
 int      dof;          // The number of degrees of freedom = SpaceDim / 2.
 complex* refPoint;     // Default reference point in phase space about which
                        //   derivatives are to be taken.
 int*     AllZeroes;    // An integer array containing zeroes.  Used to
                        //   filter the standard part of a JL variable.
 dlist    predefJL;     // A list of JL variables whose declarations precede
                        //   a call to JLSetup.  These are post-processed when
                        //   JLSetup is finally called.
 dlist    predefLieOp;  // A list of CLieOperator variables whose declarations precede
                        //   a call to JLSetup.  These are post-processed when
                        //   JLSetup is finally called.
 JetC__environment();
 ~JetC__environment();
};


class JetC {

private:
  JLC*  jl;

  JetC EpsInverse() const;  // Calculates the inverse of
                            // nilpotent Jets.

  // Friends
  friend JetC concat( const JetC& );
  friend class coordC;
  friend class LieOperator;
  friend class CLieOperator;

  static dlist  coordPool;
  static dlist  coordPtrs;

public:
  char        stacked;    // ??? Made public for convenience: no application
                          // ??? should touch this variable.

  // Constructors and destructors_____________________________________
  JetC();
  JetC( const complex );
  JetC( const JetC& );
  JetC( const Jet&);
  ~JetC();

  static void Setup( const int      = 6, /* Total number of variables */
                     const int      = 1, /* Maximum derivative weight */
                     const int      = 0, /* Dimension of phase space  */
                     const complex* = 0  /* Default reference point   */
                   );

  // Obsolete setup routines -----------------------------------------
  static void setup( const int, const int );
  static void setup( const int, const int, const complex* );
  static void setup( const int, const int, const int );
  static void setup( const int, const int, const int, const complex* );
  // ------------------------------------------------------------------

  static void FixReference        ( const complex* );
  static void FixReference        ( const int* );
  static void FixReference        ( const JetC& );
  static void FixReferenceAtStart ( const CLieOperator& );
  static void FixReferenceAtEnd   ( const CLieOperator& );

  static JetC__environment oneEnv;    // ??? Make private eventually!
  static FILE  *scratchFile;

  // Public member functions__________________________________________
  JLC* operator->() const { return jl; }

  JLCterm* get();           // Pops the top term, which should be the 
                            // one of lowest weight.
  void addTerm( JLCterm* ); // Public only for diagnostic purposes.

  void fixReference();
  void fixReference( const complex* );
  void fixReference( const JetC& );
  void fixReferenceAtEnd( const CLieOperator& );
  void fixReferenceAtStart( const CLieOperator& );

  char isNilpotent();
  void peekAt();
  void printCoeffs();
  dlist image();
  void writeToFile( char*   /* Name of unopened file */ );
  void writeToFile( FILE* );

  void getReference( complex* );
  void scaleBy( complex );
  void setVariable( const complex, const int );
  void setVariable( int );
  complex standardPart() const;
  void clear();
  complex weightedDerivative( int* );
  complex derivative( int* );
  JetC filter(  const int, const int ) const;  
                                  // Returns only those JLterms whose weight 
          	                  // are between two specified values, inclusive.
  JetC filter( char (*)( const int*, const complex ) ); 
                                   // Returns those JLterms for which the 
                                   // argument is satisfied.
  complex operator()( complex* ) const;	   
                                   // Performs a multinomial evaluation of 
				   // the JetC variable.  Essentially acts as a 
				   // power series expansion.
  JetC  operator()( JetC* ) const;   // Self explanatory ...
  JetC  operator()( CLieOperator& ) const; 
                                     // Self explanatory ...
  JetC  D( int* )const ;	     // Performs differentiation of a JetC variable.

  friend Jet real( const JetC& );
  friend Jet imag( const JetC& );

  // Operators________________________________________________________
  friend ostream& operator<<(ostream&, JetC&);
  friend istream& operator>>(istream&, JetC&);
  friend char operator==( const JetC&, const JetC& );
  friend char operator==( const JetC&, const complex& );
  friend char operator==( const complex&, const JetC& );

  JetC& operator=( const JetC& );
  JetC& operator=( const complex& );
  void operator+=( const complex& );

  friend JetC operator+( const JetC&, const JetC& );
  friend JetC operator+( const JetC&, const complex& );
  friend JetC operator+( const complex&, const JetC& );
  friend JetC operator-( const JetC&, const JetC& );
  friend JetC operator-( const JetC& );
  friend JetC operator-( const JetC&, const complex& );
  friend JetC operator-( const complex&, const JetC& );
  friend JetC operator*( const JetC&, const JetC& );
  friend JetC operator*( const JetC&, const complex );
  friend JetC operator*( const complex, const JetC& );
  friend JetC operator*( const JetC&, const double );
  friend JetC operator*( const double, const JetC& );
  friend JetC operator|( const JetC&, const JetC& );
  friend JetC operator/( const JetC&, const JetC& );
  friend JetC operator/( const JetC&, const complex& ); 
  friend JetC operator/( const JetC&, const double& ); 
  friend JetC operator/(       complex, const JetC& ); 
  friend JetC operator/(       double, const JetC& ); 
  friend JetC operator^( const JetC&, const JetC& );   // Poisson bracket

  friend JLterm operator*( JLterm&, JLterm& );
  friend JetC operator*( const JetC&, const int );
  friend JetC operator*( const int, const JetC& );
  friend JetC operator/( const JetC&, const int& ); 

  friend CLieOperator operator*( const CLieOperator&, const CLieOperator& );

  JetC inverse();            // Returns the inverse JetC
                             //  only when the number of
                             //  coordinates = 1.

  JetC Inverse() const;      // Returns the inverse jet
                             //  only when the number of
                             //  coordinates = 1.

 // Overloaded analytic functions__________________________________
  friend JetC acos( const JetC& );
  friend JetC asin( const JetC& );
  friend JetC atan( const JetC& );
  //  friend JetC atan2( const JetC&, const JetC& );
  friend JetC cos( const JetC& );
  friend JetC cosh( const JetC& );
  friend JetC exp( const JetC& );
  friend JetC log( const JetC& );
  friend JetC log10( const JetC& );
  friend JetC pow( const JetC&, const double );
  friend JetC pow( const JetC&, int );
  friend JetC sin( const JetC& );
  friend JetC sinh( const JetC& );
  friend JetC sqrt( const JetC& );
  friend JetC tan( const JetC& );
  friend JetC tanh( const JetC& );
};

JetC erf( const JetC& );
JetC w( const JetC&  );

void operator+=( JetC&, const JetC& );   // ??? Why are these void operators?
void operator-=( JetC&, const JetC& );
void operator*=( JetC&, const JetC& );
void operator/=( JetC&, const JetC& );
void operator/=( JetC&, const complex );
char operator!=( const JetC&, const JetC& );
char operator!=( const JetC&, const complex& );
char operator!=( const complex&, const JetC& );

// *******************************************************************

class coordC : public JetC {
private:
  int index;
public:
  coordC();
  coordC( const complex );
  coordC( const coordC& );
  ~coordC();
  coordC& operator=( const coordC& );
  coordC& operator=( const JetC& );
  void set( const complex );
} ;

#endif // JetC_HXX
